import copy
import json
import os
from lib2to3.pgen2.token import OP
from typing import Dict, List, Optional, Tuple, Type
from sensiml.dclproj.confusion_matrix import ConfusionMatrix
import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame, concat
from sensiml.dclproj.vizualization import plot_segments
from sensiml.dclproj.utils import upload_datasegments


class DataSegment(object):
    def __init__(
        self,
        segment_id: int,
        capture_sample_sequence_start: int,
        capture_sample_sequence_end: int,
        columns: Optional[list] = None,
        data: Optional[np.array] = None,
        session: Optional[str] = None,
        label_value: Optional[str] = None,
        uuid: Optional[str] = None,
        capture: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        **kwargs,
    ):
        self._metadata = [
            "label_value",
            "capture",
            "segment_id",
            "capture_sample_sequence_start",
            "capture_sample_sequence_end",
            "session",
            "uuid",
            "segment_length",
        ]
        self._label_value = "" if None else str(label_value)
        self._session = session
        self._segment_id = segment_id
        self._uuid = uuid
        self._capture = capture
        self._data = data
        self._columns = columns
        self._capture_sample_sequence_start = capture_sample_sequence_start
        self._capture_sample_sequence_end = capture_sample_sequence_end
        self._extra_metadata = extra_metadata

    def __repr__(self):
        return str(self.metadata)

    def copy(self):
        """Returns a deepcopy of this item

        Returns:
            DataSegment: Returns a deepcopy of the current DataSegment object
        """
        return copy.deepcopy(self)

    def to_dict(self):
        tmp = self.metadata
        tmp["columns"] = self.columns
        tmp["data"] = self.data

        return tmp

    @property
    def metadata(self):
        tmp_metadata = {k: getattr(self, f"_{k}") for k in self._metadata}
        if self._extra_metadata:
            tmp_metadata.update(self._extra_metadata)

        return tmp_metadata

    @property
    def capture(self):
        return self._capture

    @property
    def segment_id(self):
        return self._segment_id

    @property
    def label_value(self):
        return self._label_value

    @property
    def uuid(self):
        return str(self._uuid)

    @property
    def has_data(self):
        if self._data is not None:
            return True

        return False

    @property
    def data(self):
        return self._data

    @property
    def segment_length(self):
        return self._capture_sample_sequence_end - self._capture_sample_sequence_start

    @property
    def _segment_length(self):
        return self.segment_length

    @property
    def columns(self):
        return self._columns

    @property
    def capture_sample_sequence_start(self):
        return self._capture_sample_sequence_start

    @property
    def capture_sample_sequence_end(self):
        return self._capture_sample_sequence_end

    @property
    def start(self):
        return self.capture_sample_sequence_start

    @property
    def end(self):
        return self.capture_sample_sequence_end

    @start.setter
    def start(self, value):
        self.capture_sample_sequence_start = value

    @end.setter
    def end(self, value):
        self.capture_sample_sequence_end = value

    @label_value.setter
    def label_value(self, value):
        if value is None:
            self._label_value = "Unknown"
        else:
            self._label_value = value

    @property
    def session(self):
        return self._session

    def to_dataframe(self):
        if self.data is None:
            return

        return DataFrame(self.data.T, columns=self.columns)

    def get_column_index(self, column):
        if self._columns is None:
            return

        return self._columns.index(column)

    def plot(self, figsize: Tuple = (30, 4), currentAxis=None, **kwargs):

        if self._data is None:
            return

        if currentAxis is None:
            plt.figure(figsize=figsize)
            currentAxis = plt.gca()

        for index, axis in enumerate(self.data):
            currentAxis.plot(axis, label=self.columns[index])

        currentAxis.set_xlim(0, self.data.shape[1])

        plt.legend(loc=0)

        return currentAxis

    def plot_spectrogram(
        self, channel: str, fft_length: int = 512, figsize: Tuple = (30, 4), **kwargs
    ):
        """Plots the spectrogram for the signal.

        Args:
            channel (str): the channel/column of sensor data to use
            fft_length (int, optional): The size of the FTT length to use when computing the spectrogram. Defaults to 512.
            figsize (Tuple, optional): the size of the figure that will be created. Defaults to (30, 4).
        """
        if self._data is None:
            return

        fig, axis = plt.subplots(figsize=figsize)

        data = axis.specgram(
            self.data[self.get_column_index(channel)], NFFT=fft_length, **kwargs
        )


class DataSegments(object):
    def __init__(self, data: Optional[list] = None):

        self._data = []

        if data:
            self._data = data

    @property
    def data(self):
        return self._data

    def __repr__(self):
        display(self.to_dataframe())
        return ""

    def __str__(self):
        return self.to_dataframe().__str__()

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __add__(self, value):
        return DataSegments(self._data + value._data)

    def __getitem__(self, key):
        return self._data[key]

    def head(self, value=5):
        return self.to_dataframe().head(value)

    def append(self, data_segment: DataSegment):
        """Append a DataSegment to the current DataSegments List

        Args:
            data_segment (DataSegment):
        """
        self._data.append(data_segment)

    def extend(self, data_segments: DataSegment):
        """Extend the current DataSegments object with another DataSegments object

        Args:
            data_segments (DataSegment):
        """

        self._data.extend(data_segments._data)

    @property
    def only_metadata(self):
        if self._data is None:
            return True

        if self._data[0].data is None:
            return True

        return False

    def to_dataframe(self, metadata_only: bool = True):
        M = []

        if not self._data:
            return None

        if self.only_metadata:
            for segment in self._data:
                M.append(segment.metadata)

            return DataFrame(M)

        for segment in self.data:
            if metadata_only:
                M.append(DataFrame([segment.metadata]))
            else:
                tmp_df = DataFrame(segment.data.T, columns=segment.columns)
                M.append(tmp_df.assign(**segment.metadata))

        return concat(M).reset_index(drop=True)

    def apply(self, func, **kwargs) -> DataFrame:
        """Apply a function to all the segments in the DataSegments object and return a DataFrame of the resulting generated features for each datasegment.

        Args:
            func (_type_): and function object which takes a DataSegment as its first input and kwargs as the following

        Returns:
            DataFrame: A DataFrame of the generated features from the applied function
        """

        feature_vectors = []
        for segment in self.data:
            tmp_df = func(segment, **kwargs)
            feature_vectors.append(tmp_df.assign(**segment["metadata"]))

        return concat(feature_vectors).reset_index(drop=True)

    def iter_dataframe(self):

        if self.only_metadata:
            for segment in self._data:
                yield DataFrame([segment["metadata"]])
        else:
            for segment in self._data:
                tmp_df = DataFrame(segment["data"].T, columns=segment["columns"])

                yield tmp_df.assign(**segment["metadata"])

    def iterrows(self):
        for index, data_segment in enumerate(self.data):
            yield (index, data_segment.to_dict())

    def to_dcli(
        self,
        filename: Optional[str] = None,
        session: Optional[str] = None,
        verification_id: Optional[str] = None,
        session_parameters: Optional[dict] = None,
    ) -> List:
        """Creates a .dcli file describing the segment information that can be imported into the Data Capture Lab

        Args:
            filename (Optional[str], optional): The name of the file to save it to, if None no file is created.. Defaults to None.
            session (Optional[str], optional): The name of a session to use when creating the DCLI file. if None the session from the DataSegment objects are used.. Defaults to None.

        Returns:
            List: DCLI formatted segments
        """

        def get_capture_index(dcli_capture, file_name):
            for index, capture in enumerate(dcli_capture):
                if file_name == capture["file_name"]:
                    return index

            return None

        def get_session_index(sessions, session_name):
            for index, session in enumerate(sessions):
                if session_name == session["session_name"]:
                    return index

            return None

        dcli_capture = []

        for segment in self._data:

            capture_index = get_capture_index(dcli_capture, segment.capture)

            if capture_index is None:
                dcli_capture.append({"file_name": segment.capture, "sessions": []})
                capture_index = -1

            session_index = get_session_index(
                dcli_capture[capture_index]["sessions"],
                segment.session if session is None else session,
            )

            if session_index is None:
                if session is None:
                    dcli_capture[capture_index]["sessions"].extend(
                        [{"session_name": segment.session, "segments": []}]
                    )
                else:
                    dcli_capture[capture_index]["sessions"].extend(
                        [{"session_name": session, "segments": []}]
                    )

                if verification_id:
                    dcli_capture[capture_index]["sessions"][-1][
                        "verification_id"
                    ] = verification_id
                if session_parameters:
                    dcli_capture[capture_index]["sessions"][-1][
                        "session_parameters"
                    ] = session_parameters

                session_index = -1

            dcli_capture[capture_index]["sessions"][session_index]["segments"].append(
                {
                    "name": "Label",
                    "value": segment.label_value,
                    "start": segment.start,
                    "end": segment.end,
                }
            )

        if filename is not None:
            print("writing dcli file to", filename)
            json.dump(dcli_capture, open(filename, "w"))

        return dcli_capture

    def to_audacity(self, rate: int = 16000) -> List:
        """Creates multiple files with the naming convention file_{capture_name}session{session_name}.txt.

        These can be imported into Audacity directly going to File->Import->Labels

        Args:
            rate (int): Audacity uses the actual time and note number of samples. Set the rate to the sample rate for the captured date. Default is 16000.
        """

        dcli = self.to_dcli()

        for capture in dcli:
            for session in capture["sessions"]:
                outfile = "file_{capture}_session_{session}.txt".format(
                    capture=capture["file_name"], session=session["session_name"]
                )
                with open(
                    outfile,
                    "w",
                ) as out:
                    for segment in session["segments"]:
                        out.write(
                            "{start_time}\t{end_time}\t{label}\n".format(
                                start_time=segment["start"] / rate,
                                end_time=segment["end"] / rate,
                                label=segment["value"],
                            )
                        )
                    print("labels written to {outfile}".format(outfile=outfile))

    @property
    def label_values(self) -> List:
        """List all the labels in the DataSegments object"""

        label_values = set()
        for segment in self._data:
            label_values.add(segment.label_value)

        return sorted(list(label_values))

    def merge_label_values(self, data_segments: dict) -> List:
        """Merges label values between to data segments.

        Args:
            data_segments (dict): The datasegment object to merge label values with

        Returns:
            List: The sorted union of the label values from both datasegments
        """

        return sorted(
            list(set(data_segments.label_values).union(set(self.label_values)))
        )

    def merge_segments(self, delta: int = 1, verbose=False):
        """Merge segments that overlap or are within delta of each other.

        Args:
            delta (int, optional): The distance between two nonoverlapping segments where they will still be merged. Defaults to 1.

        Returns:
            DataFrame: A DataFrame consisting of the merged segments
        """

        return merge_segments(self, delta=delta, verbose=verbose)

    def filter_segments(self, min_length: int = 10000):
        """Merges data segments that are within a distance delta of each other and have the same class name."""

        new_datasegments = DataSegments()

        for _, segment in enumerate(self.data):
            if segment.segment_length > min_length:
                new_datasegments.append(segment)

        print(
            "Original Segments:",
            len(self),
            "Filtered Segments:",
            len(new_datasegments),
        )

        return new_datasegments

    def remove_overlap(self, verbose: bool = False, inplace: bool = False):
        """Removes the overlap between segments by setting the segment start and
        end of overlapping segments to the same point, halfway between the overlapping edges.

        Args:
            dcl (DCLProject): A DCLProject object that is connected to the DCLI file

        Returns:
            DataSegments: A DataSegments object consisting of the merged segments
        """
        return remove_overlap(self, verbose=verbose, inplace=inplace)

    def join_segments(self, delta: Optional[int] = None, inplace: bool = False):
        """Joins adjacent segments so that there is no empty space between segments.

        Args:
            dcl (DCLProject): A DCLProject object that is connected to the DCLI file
            delta (int): Segments outside this range will not be joined. If None, all neighboring segments will be merged regardless of the distance. Default is None.

        Returns:
            DataSegments: A DataSegments object consisting of the merged segments
        """

        return join_segments(self, delta=delta, inplace=inplace)

    def plot_segments(
        self,
        capture: Optional[DataFrame] = None,
        capture_name: Optional[str] = None,
        labels: Optional[list] = None,
        **kwargs,
    ):
        """plot the datasegment labels

        Args:
            capture (DataFrame, optional): DataFrame of capture data. Defaults to None.
            capture_name (str, optional): name of the capture. Defaults to None.
            labels (List[str], optional): List of labels to use. Defaults to None.
            kwargs
        """

        plot_segments(
            self, capture=capture, capture_name=capture_name, labels=labels, **kwargs
        )

    def nearest_labels(
        self,
        ground_truth_segments,
        overlap_pct: float = 0.5,
        verbose=False,
        keep_default=False,
    ) -> dict:
        """Computes the nearest labels in the current DataSegment to a ground truth DataSegments and updates the labels with
        the ground truth label.

        Args:
            pred_segments (list): list of segments from prediction file
            ground_truth_segments (list): list of segments from manually labeled file

        Returns:
            DataSegment: A DataSegments object with the new updated labels
        """
        tmp_segments = copy.deepcopy(self)
        # Loop over all predicted segments
        for new_seg in tmp_segments:
            # Loop over real segments to find all real segments that the predicted segment overlaps with
            # (discard predictions that overlap w/ nothing)
            last_overlap = 0
            for gt_seg in ground_truth_segments:
                if gt_seg.segment_length == 0:
                    break
                # If the next real segment begins after the predicted one ends then there will be no more overlapping segments
                if gt_seg.start > new_seg.end:
                    break
                # If the next real segment ends before the predicted one starts then it won't overlap so we should skip it
                if gt_seg.end < new_seg.start:
                    continue
                overlap = _calculate_overlap_percent(gt_seg, new_seg)
                # Keep track of what the prediction was and how it compared to the actual label IF it is over the overlap threshold
                if overlap >= overlap_pct:
                    if overlap > last_overlap:
                        if verbose and new_seg.label_value != gt_seg.label_value:
                            print(
                                f"Updating {new_seg.segment_id} from {new_seg.label_value} to {gt_seg.label_value}"
                            )
                        new_seg.label_value = gt_seg.label_value
                        last_overlap = overlap

            if last_overlap == 0:
                if verbose and not keep_default:
                    print(
                        f"No matching label found for {new_seg.segment_id} setting to None"
                    )
                if not keep_default:
                    new_seg.label_value = None

        return tmp_segments

    def confusion_matrix(
        self, ground_truth: dict, overlap_pct: float = 0.5
    ) -> ConfusionMatrix:
        """Generate confusion matrix with this segments and overlapping segments

        Args:
            ground_truth (DataSegments): DataSegments to use as the ground truth
            overlap_pct (float, optional): amount of overlap to consider the segments overlapping. Defaults to 0.5.

        Returns:
            ConfusionMatrix: confusion matrix object
        """
        return ConfusionMatrix(self, ground_truth, overlap_percent=overlap_pct)

    def upload(self, client, default_label: str, verbose: bool = True):
        """Upload the segments

        Args:
            client (client): Client logged in and connected to the target project
            default_label (str): default label to use
            verbose (bool, optional): prints out info. Defaults to True.
        """

        upload_datasegments(client, self, default_label=default_label, verbose=verbose)


def _get_percent(numerator: int, denominator: int) -> float:

    return round(100 * numerator / denominator, 2) if denominator and numerator else 0


def _calculate_overlap_percent(s1, s2) -> float:

    # calculate overlap between labels
    label_overlap = min(s1.end, s2.end) - max(s1.start, s2.start)
    # calculate length of label 2
    range_of_label_2 = s2.end - s2.start
    # return overlap percent
    return _get_percent(label_overlap, range_of_label_2)


def check_overlap(s1, s2):
    return (
        True
        if max(
            0,
            min(s1.end, s2.end) - max(s1.start, s2.start),
        )
        > 0
        else False
    )


def check_near(s1, s2, delta):
    if delta is None:
        return True

    return True if abs(s2.start - s1.end) < delta else False


def merge_segments(
    segments: DataSegments, delta: int = 10, verbose=False
) -> DataSegments:
    """Merges data segments that are within a distance delta of each other and have the same class name.

    Args:
        segments (DataFrame): A DataFrame of segments
        delta (int, optional): The distance between two nonoverlapping segments where they will still be merged. Defaults to 10.

    Returns:
        DataFrame: DataFrame containing the merged segments
    """

    seg_groups = segments.to_dataframe().groupby(["session", "capture"])

    new_segments = DataSegments()
    for key in seg_groups.groups.keys():
        if verbose:
            print("Group", key)

        segment_list = (
            seg_groups.get_group(key)
            .sort_values(by="capture_sample_sequence_start")
            .index.values.tolist()
        )
        merge_list = []

        for index, segment_index in enumerate(segment_list):

            segment = segments[segment_index]
            next_segment = None

            if len(segment_list) - 1 != index:
                next_segment = segments[segment_list[index + 1]]

            if not merge_list:
                if len(segment_list) - 1 == index:
                    new_segments.append(segment)
                    continue
                elif next_segment.label_value != segment.label_value:
                    new_segments.append(segment)
                    continue
                elif not check_near(segment, next_segment, delta) and not check_overlap(
                    segment, next_segment
                ):
                    new_segments.append(segment)
                    continue

            if (
                len(segment_list) - 1 != index
                and next_segment.label_value == segment.label_value
                and (
                    check_near(segment, next_segment, delta)
                    or check_overlap(segment, next_segment)
                )
            ):

                if not merge_list:
                    merge_list.append(index)
                merge_list.append(index + 1)

            else:
                # do merge
                if merge_list:
                    if verbose:
                        print("merging", merge_list)

                    new_segments.append(merge_segment(segments, merge_list))
                else:
                    new_segments.append(segment)

                merge_list = []

        if merge_list:
            if verbose:
                print("merging", merge_list)
            new_segments.append(merge_segment(segments, merge_list))

    print("Original Segments:", len(segments), "Merged Segments:", len(new_segments))

    return new_segments


def merge_segment(segments, merge_list):

    tmp_segment = segments[merge_list[0]].copy()

    end = tmp_segment.end

    for segment_index in merge_list:
        if end < segments[segment_index].end:
            end = segments[segment_index].end

    tmp_segment._capture_sample_sequence_end = end

    if tmp_segment.has_data:
        capature_start = tmp_segment.start
        data = np.zeros((tmp_segment.data.shape[0], end - tmp_segment.start))
        head_index = 0
        capture_head_index = capature_start
        for segment in [segments[x] for x in merge_list]:
            segment.start + head_index
            data[:, head_index : segment.end - capature_start] = segment.data[
                :, capture_head_index - segment.start : segment.end - segment.start
            ]
            head_index += segment.end - capature_start
            capture_head_index += head_index

        tmp_segment._data = data

    return tmp_segment


def remove_overlap(
    segments: DataSegments, verbose: bool = False, inplace: bool = False
) -> DataSegments:
    """Removes the overlap between segments by setting the segment start and end of
    overlapping segments to the same point, halfway between the overlapping edges."""

    if inplace:
        new_segments = segments
    else:
        new_segments = copy.deepcopy(segments)

    seg_groups = new_segments.to_dataframe().groupby(["session", "capture"])

    for key in seg_groups.groups.keys():
        if verbose:
            print("Group", key)

        segment_list = (
            seg_groups.get_group(key)
            .sort_values(by="capture_sample_sequence_start")
            .index.values.tolist()
        )

        for index, segment_index in enumerate(segment_list[:-1]):
            segment = new_segments[segment_index]
            next_segment = new_segments[segment_list[index + 1]]

            if check_overlap(segment, next_segment):
                difference = (abs((next_segment.start - segment.end)) + 1) // 2
                segment._capture_sample_sequence_end -= difference
                next_segment._capture_sample_sequence_start += difference

                if segment.has_data:
                    segment._data = segment._data[:, :-difference]
                if next_segment.has_data:
                    next_segment._data = segment._data[:, difference:]

    return new_segments


def join_segments(
    segments, delta: Optional[int] = None, inplace: bool = False
) -> DataFrame:
    """If there are any gaps between two segments, this will bring them together so there are no unlabeled regions of data."""

    print(
        "This will drop the data, use the .refresh_data API to repopulate the data table"
    )

    if inplace:
        new_segments = segments
    else:
        new_segments = copy.deepcopy(segments)

    seg_groups = new_segments.to_dataframe().groupby(["session", "capture"])

    for key in seg_groups.groups.keys():

        segment_list = (
            seg_groups.get_group(key)
            .sort_values(by="capture_sample_sequence_start")
            .index.values.tolist()
        )

        for index, segment_index in enumerate(segment_list[:-1]):
            segment = new_segments[segment_index]
            next_segment = new_segments[segment_list[index + 1]]

            if (
                not check_overlap(segment, next_segment) and True
                if delta is None
                else check_near(segment, next_segment, delta=delta)
            ):
                difference = (abs((next_segment.start - segment.end)) + 1) // 2
                segment._capture_sample_sequence_end += difference
                next_segment._capture_sample_sequence_start -= difference

        segment._data = None
        next_segment._data = None

    return new_segments


def model_results_to_datasegments(
    results: DataFrame,
    data: Optional[np.array] = None,
    capture_name: Optional[str] = None,
    session: Optional[str] = None,
):
    datasegments = DataSegments()
    for index, result in enumerate(results.to_dict(orient="records")):
        if result["capture_sample_sequence_start"] < 0:
            print("Filtering: capture_sample_sequence_start must be > 0", result)
            continue
        if (
            result["capture_sample_sequence_start"]
            > result["capture_sample_sequence_end"]
        ):
            print(
                "Filtering: capture_sample_sequence_start must be < capture_sample_sequence_end",
                result,
            )

        if data is None:
            datasegments.append(
                DataSegment(
                    segment_id=index,
                    data=None,
                    session=session,
                    capture=capture_name,
                    **result,
                )
            )
        else:
            datasegments.append(
                DataSegment(
                    segment_id=index,
                    columns=data.columns.to_list(),
                    data=data.values[
                        result["capture_sample_sequence_start"] : result[
                            "capture_sample_sequence_end"
                        ]
                    ].T,
                    session=session,
                    capture=capture_name,
                    **result,
                )
            )

    return datasegments


def segment_list_to_datasegments(
    labels: DataFrame, session: str = "", dcl: Optional[object] = None
):
    """Converts a DataFrame of segments into a DataSegments object

    Args:
        labels (DataFrame): A dataframe containing the segment information
        session (str, optional): The session to set the segments too. Defaults to "".
        dcl (DCLProject): A DCLProject object that is connected to the DCLI file, If this is passed in the data property of the DataSegment objects will be filled with sensor data

    Returns:
        DataSegments
    """

    if isinstance(labels, DataFrame):
        # DCL generates segments with this format ""
        if labels.columns.to_list() == [
            "File",
            "Label",
            "Start",
            "End",
            "Length",
        ]:
            label_dict = (
                labels.rename(
                    columns={
                        "File": "capture",
                        "Label": "label_value",
                        "Start": "capture_sample_sequence_start",
                        "End": "capture_sample_sequence_end",
                    },
                )
                .sort_values(by="capture")
                .to_dict(orient="records")
            )
        elif "Seg_Begin" in labels.columns.to_list():
            if "capture" not in labels.columns:
                labels["capture"] = None
            if "label_value" not in labels.columns:
                labels["label_value"] = None
            label_dict = (
                labels.rename(
                    columns={
                        "Seg_Begin": "capture_sample_sequence_start",
                        "Seg_End": "capture_sample_sequence_end",
                    },
                )
                .sort_values(by="SegmentID")
                .to_dict(orient="records")
            )
        else:
            if "segmenter" in labels.columns.tolist():
                labels.rename({"segmenter": "session"}, axis=1, inplace=True)
            label_dict = labels.sort_values(
                by=["capture", "session", "capture_sample_sequence_start"]
            ).to_dict(orient="records")
    else:
        raise Exception("Expected DataFrame")

    data_segments = DataSegments()
    capture_name = None
    columns = None
    data = None
    capture = None

    for index, label in enumerate(label_dict):

        if capture_name != label["capture"]:
            if dcl:
                capture = dcl.get_capture(label["capture"])
            else:
                capture = None
            capture_name = label["capture"]

        if capture:
            data = capture.iloc[
                label["capture_sample_sequence_start"] : label[
                    "capture_sample_sequence_end"
                ]
            ].values.T
            columns = capture.columns.values.tolist()
        else:
            data = None
            columns = None

        data_segments.append(
            DataSegment(
                segment_id=index,
                capture_sample_sequence_start=label["capture_sample_sequence_start"],
                capture_sample_sequence_end=label["capture_sample_sequence_end"],
                data=data,
                columns=columns,
                session=label.get("session", session),
                label_value=label["label_value"],
                capture=label["capture"],
            )
        )

    return data_segments


def audacity_to_datasegments(
    capture_name, file_path: str, session: str = "", rate: int = 16000
):
    """Converts labels exported from Audacity into a datasegment object.

    Args:
        capture_name (str): The name of the capture file to import
        file_path (DataFrame): The file path to the Audacity Label
        session (str, optional): The session to set the segments too. Defaults to "".
        rate (int): Audacity uses the actual time and note number of samples. Set the rate to the sample rate for the captured date. Default is 16000.
        data (DataFrame): The data associated with the audacity labels


    Returns:
        Datasegments
    """

    M = []
    with open(file_path, "r") as fid:
        for line in fid.readlines():
            if line[0] == "\\":
                continue
            start, end, label = line[:-1].split("\t")
            M.append([float(start), float(end), label])

    segment_list = DataFrame(
        M,
        columns=[
            "capture_sample_sequence_start",
            "capture_sample_sequence_end",
            "label_value",
        ],
    )

    segment_list["capture_sample_sequence_start"] = (
        segment_list["capture_sample_sequence_start"] * rate
    ).astype(int)
    segment_list["capture_sample_sequence_end"] = (
        segment_list["capture_sample_sequence_end"] * rate
    ).astype(int)

    segment_list["capture"] = capture_name
    segment_list["session"] = session

    return segment_list_to_datasegments(segment_list)



import json
import datetime
from pandas import DataFrame
import re
import warnings

import sensiml.base.utility as utility


class QueryColumns(object):
    """Base class for the query columns and metadata_columns properties"""

    def __init__(self, columns=[]):
        if len(columns) > 0:
            self._columns = columns
        else:
            self._columns = []

    def add(self, *items):
        """Adds a column name or names for query selection.

        Args:
            items (str or list[str]): column names to add

        Note:
            The named column(s) must exist in the project and duplicate column names will be ignored
        """
        for item in items:
            if item not in self._columns:
                self._columns.append(item)

    def remove(self, item):
        """Removes a column name from query selection.

        Args:
            item (str): column name to remove
        """
        if item in self._columns:
            self._columns.remove(item)

    def clear(self):
        """Clears all column names from query selection"""
        self._columns = []

    def __str__(self):
        return json.dumps(self._columns)

    def __iter__(self):
        return iter(self._columns)


class Query(object):
    """Base class for a query.

    Queries extract project data, or a subset of project data, for use in a pipeline. The query must specify which
    columns of data to extract and what filter conditions to apply.
    """

    def __init__(self, connection, project):
        """Initializes a query instance"""
        self._uuid = ""
        self._name = ""
        self._columns = QueryColumns()
        self._metadata_columns = QueryColumns()
        self._metadata_filter = ""
        self._segmenter = None
        self._label_column = ""
        self._combine_labels = None
        self._dirty = False
        self._capture_configurations = ""
        self._connection = connection
        self._summary_statistics = None
        self._project = project
        self._created_at = None
        self._cache = None
        self._task_status = None
        self._last_modified = None

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """Name of the query"""
        return self._name

    @property
    def summary_statistics(self):
        """Name of the query"""
        return self._summary_statistics

    @summary_statistics.setter
    def summary_statistics(self, value):
        self._summary_statistics = value

    @name.setter
    def name(self, value):
        self._dirty = True
        self._name = value

    @property
    def columns(self):
        """Sensor columns to include in the query result

        Note:
            Columns must correspond to actual project sensor columns or the reserved word 'SequenceID' for the
            original sample index.
        """
        self._dirty = True
        return self._columns

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        self._created_at = datetime.datetime.strptime(value[:19], "%Y-%m-%dT%H:%M:%S")

    @property
    def label_column(self):
        """Label columns to use in the query result

        Note:
            Columns must correspond to actual project label column"""

        self._dirty = True
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        self._dirty = True
        self._label_column = value

    @property
    def combine_labels(self):
        """Combine label values into new value to use in the query result

        Label = Gesture
        Label_Values = A,B,C,D,E
        combine_labels = {'Group1':['A','B',C'],'Group2':['D','E']}

        the labels that will be returned will be group1 and group2
        """

        self._dirty = True
        return self._combine_labels

    @combine_labels.setter
    def combine_labels(self, value):
        self._dirty = True
        self._combine_labels = value

    @property
    def metadata_columns(self):
        """Metadata columns to include in the query result

        Note:
            Columns must correspond to actual project metadata columns.
        """
        self._dirty = True
        return self._metadata_columns

    @property
    def metadata_filter(self):
        """Filter criteria of the query

        Args:
            value (str): similar to a SQL WHERE clause, the string can contain any number of AND-concatenated
              expressions where square brackets surround the column name and comparison value, with the operator in
              between. Supported operators: >, >=, <, <=, =, !=, IN

        Examples::

            metadata_filter = '[Subject] > [5] AND [Subject] <= [15]'
            metadata_filter = '[Gender] = [Female] AND [Activity] != [Walking]'
            metadata_filter = '[Subject] IN [5, 7, 9, 11, 13, 15]'

        Note:
            Queries do not support OR-concatenation between expressions, but often the IN operator can be used to
            achieve OR-like functionality on a single column. For example::

                [Gesture] IN [A, M, L]

            is equivalent to::

                [Gesture] = [A] OR [Gesture] = [M] OR [Gesture] = [L]

        """
        return self._metadata_filter

    @metadata_filter.setter
    def metadata_filter(self, value):
        self._dirty = True
        self._metadata_filter = value

    @property
    def capture_configurations(self):
        return self._capture_configurations

    @capture_configurations.setter
    def capture_configurations(self, value):
        self._dirty = True

        if isinstance(value, str):
            value = [value]

        if not isinstance(value, list):
            raise Exception("capture_configurations must be List or String")

        self._capture_configurations = json.dumps(value)

    @property
    def segmenter(self):
        """Segmenter to use for the query

        Args:
            value (int): ID of segmenter.

        """
        return self._segmenter

    @segmenter.setter
    def segmenter(self, value):
        if value is None or isinstance(value, int):
            self._segmenter = value
        else:
            raise ValueError("Value must be a integer")

    @property
    def task_status(self):
        return self._task_status

    @task_status.setter
    def task_status(self, value):
        self._task_status = value

    @property
    def last_modified(self):
        return self._last_modified

    @last_modified.setter
    def last_modified(self, value):
        self._last_modified = datetime.datetime.strptime(
            value[:19], "%Y-%m-%dT%H:%M:%S"
        )

    @property
    def cache(self):
        return self._cache

    @cache.setter
    def cache(self, value):
        self._cache = value

    def insert(self, renderer=None):
        """Calls the REST API and inserts a new query."""

        url = "project/{0}/query/".format(self._project.uuid)

        data = {
            "name": self.name,
            "columns": self.columns._columns,
            "metadata_columns": self.metadata_columns._columns,
            "metadata_filter": self.metadata_filter,
            "segmenter_id": self._segmenter,
            "label_column": self.label_column,
            "combine_labels": self.combine_labels,
            "capture_configurations": self.capture_configurations,
        }

        request = self._connection.request("post", url, data)
        response, err = utility.check_server_response(request, renderer=renderer)
        if err is False:
            self.uuid = response["uuid"]
            self._dirty = False

    def update(self, renderer=None):
        """Calls the REST API and updates the query object on the server."""

        url = "project/{0}/query/{1}/".format(self._project.uuid, self.uuid)

        self._checked = None
        data = {
            "name": self.name,
            "columns": self.columns._columns,
            "metadata_columns": self.metadata_columns._columns,
            "metadata_filter": self.metadata_filter,
            "segmenter_id": self._segmenter,
            "label_column": self.label_column,
            "combine_labels": self.combine_labels,
            "capture_configurations": self.capture_configurations,
        }

        if len(data["columns"]) > 0 or len(data["metadata_columns"]) > 0:
            filters = (
                self._metadata_filter.split("AND")
                if len(self._metadata_filter) > 0
                else []
            )

            for i, filter in enumerate([filter for filter in filters if len(filters)]):
                match = re.search(
                    "\[(?P<key>[0-9A-Za-z_\-\. \)\(]+)\]([ ]*(?P<symbol>[\>\=|\<\=|\=|\<\>|\>|\<|in|IN|!=]+)[ ]*\[(?P<value>.+)\])?",
                    str(filter),
                )
                if not match:
                    self._checked = 1
            if self._checked is None:
                request = self._connection.request("put", url, data)
                response, err = utility.check_server_response(
                    request, renderer=renderer
                )
            else:
                self._metadata_filter = ""
                print("Metadata Filter is not formatted correctly!")
        else:
            print(
                "Sensor columns and, or metadata columns must be specified in order to query!"
            )

    def delete(self, renderer=None):
        """Calls the REST API and deletes the query object from the server."""
        url = "project/{0}/query/{1}/".format(self._project.uuid, self.uuid)
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(request, renderer=renderer)
        self._dirty = False
        if err is False:
            print(response["result"])

    def cache_query(self, renderer=None):
        """Caches the current version of the query."""
        url = "project/{0}/query/{1}/cache/".format(self._project.uuid, self.uuid)
        request = self._connection.request("post", url)
        response_data, err = utility.check_server_response(request, renderer=renderer)

        return response_data

    def cache_queryv1(self, renderer=None):
        """Caches the current version of the query."""
        url = "project/{0}/query/{1}/cache-v1/".format(self._project.uuid, self.uuid)
        request = self._connection.request("post", url)
        response_data, err = utility.check_server_response(request, renderer=renderer)

        return response_data

    def cache_queryv2(self, renderer=None):
        """Caches the current version of the query."""
        url = "project/{0}/query/{1}/cache-v2/".format(self._project.uuid, self.uuid)
        request = self._connection.request("post", url)
        response_data, err = utility.check_server_response(request, renderer=renderer)

        return response_data

    def cache_query_status(self, renderer=None):
        """Gest the status of the current caching of the query"""
        url = "project/{0}/query/{1}/cache/".format(self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(request, renderer=renderer)

        return response_data

    def cache_query_stop(self, renderer=None):
        """Kills the job for the currently executing query"""
        url = "project/{0}/query/{1}/cache/".format(self._project.uuid, self.uuid)
        request = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(request, renderer=renderer)

        return response_data

    def statistics_segments(self, renderer=None):
        """Returns metadata statistics for the query."""

        url = "project/{0}/query/{1}/statistics/".format(self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(request, renderer=renderer)
        if err is False:
            return DataFrame(response_data)
        else:
            return None

    def get_statistics_summary(self, renderer=None):
        """Returns metadata statistics for the query."""

        url = "project/{0}/query/{1}/summary-statistics/".format(
            self._project.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response, renderer=renderer)
        if err is False:
            return response_data
        else:
            return response

    def check_query_cache_up_to_date(self, renderer=None):
        """Checks if the current cached query is up to date with the current training data.

        The sensor data in a query is cached when the query is built. If the segments or metadata
        have changed since the last time the sensor data was cached then in order for a query to
        use the new data it needs to be rebuilt. This API returns whether or not the sensor data
        has changed since the last time the query was cached.

        """

        url = "project/{0}/query/{1}/cache-status/".format(
            self._project.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response, renderer=renderer)
        if err is False:
            return response_data
        else:
            return response

    def plot_statistics(self, renderer=None, **kwargs):
        """Generates a bar plot of the query statistics"""
        data = self.statistics_segments(renderer=renderer)
        if data is not None:
            data.groupby(self.label_column).size().plot(kind="bar")

    def data(self, partition=0):
        """Calls the REST API for query execution and returns the result.

        Note:
            Intended for previewing the query result before creating a query call object and using it in a sandbox
            step. The resulting DataFrame is not cached on the server, but once it is used in a sandbox it may be
            cached.
        """
        warnings.warn("This call has a timeout of two minutes.")

        if self._dirty:
            self.update()

        url = "project/{project_uuid}/query/{query_uuid}/data/{partition}".format(
            project_uuid=self._project.uuid, query_uuid=self.uuid, partition=partition
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            error_report = None
            if not isinstance(response_data, list):
                data = DataFrame(response_data)
            else:
                try:
                    data = DataFrame(response_data[0])
                    error_report = response_data[1]
                except:
                    data = DataFrame(response_data)

            return data.sort_index(axis=0), error_report
        else:
            return response

    def size(self):
        """Returns the size of the dataframe which would result from the query."""
        # Sync the client-side query with server-side query object before we check the size
        if self._dirty:
            self.update()

        url = "project/{0}/query/{1}/size/".format(self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response_data, err = utility.check_server_response(request)
        if not err:
            query_result = response_data
        else:
            query_result = None
        return query_result

    def refresh(self):
        """Calls the REST API and self populate using the uuid."""
        url = "project/{0}/query/{1}/".format(self._project.uuid, self.uuid)
        request = self._connection.request("get", url)
        response, err = utility.check_server_response(request)

        if err is False:
            self.uuid = response["uuid"]
            self.name = response["name"]
            self._columns = QueryColumns(response["columns"])
            self._label_column = response["label_column"]
            self._metadata_columns = QueryColumns(response["metadata_columns"])
            self.metadata_filter = response["metadata_filter"]
            self.segmenter = response["segmenter_id"]
            self.combine_labels = response["combine_labels"]
            self.capture_configurations = response["capture_configurations"]
            self.cache = response["cache"]
            self.task_status = response["task_status"]
            self.last_modified = response["last_modified"]
            self.summary_statistics = response["summary_statistics"]
            self._dirty = False

    def initialize_from_dict(self, data):
        """Reads a json dict and populates a single query."""
        self.uuid = data["uuid"]
        self.name = data["name"]
        self.created_at = data["created_at"]
        self.last_modified = data["last_modified"]
        self._columns = QueryColumns(data["columns"])
        self._metadata_columns = QueryColumns(data["metadata_columns"])
        self.metadata_filter = data["metadata_filter"]
        self.label_column = data.get("label_column", "")
        self.segmenter = data.get("segmenter_id", None)
        self.combine_labels = data.get("combine_labels", None)
        self.capture_configurations = data.get("capture_configurations", "")
        self.cache = data.get("cache", None)
        self.task_status = data.get("task_status", None)
        self.summary_statistics = data.get("summary_statistics")

        self._dirty = False

    def post_feature_statistics(self, window_size=None):
        """Returns metadata statistics for the query."""
        data = {"window_size": window_size}
        url = "project/{0}/query/{1}/featurestatistics/".format(
            self._project.uuid, self._uuid
        )
        response = self._connection.request("post", url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            return response_data

    def get_feature_statistics(self):
        """Returns metadata statistics for the query."""
        url = "project/{0}/query/{1}/featurestatistics/".format(
            self._project.uuid, self._uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        if err is False:
            return response_data

    def __str__(self):
        return (
            "NAME: {0}\n"
            "COLUMNS: {1}\n"
            "METADATA_COLUMNS: {2}\n"
            "METADATA FILTER: {3}\n"
            "SEGMENTER ID: {4}\n"
            "LABEL COLUMN {5}\n"
            "COMBINE LABELS {6}\n"
            "CAPTURE CONFIGURATIONS {7}"
        ).format(
            self.name,
            self.columns,
            self.metadata_columns,
            self.metadata_filter,
            self.segmenter,
            self.label_column,
            self.combine_labels,
            self.capture_configurations,
        )

    def _to_dict(self):
        return {
            "name": self.name,
            "sensor_columns": self.columns._columns,
            "metadata_columns": self.metadata_columns._columns,
            "metadata_filter": self.metadata_filter,
            "segmenter_id": self.segmenter,
            "label_column": self.label_column,
            "capture_configurations": self.capture_configurations,
        }



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter
import matplotlib.patches as patches
from copy import deepcopy
from seaborn import boxplot, violinplot, set_style
from pandas import DataFrame, concat
import io


class Visualize(object):
    def __init__(self, feature_vector=None, model=None, label=None):
        """
        model: model result or knowledgepack
        feature_vector (Dataframe): feature set that is results of generator_set or feature_selector
        label (String), name of class
        """
        self._feature_vector = None
        self._knowledgepack = None
        self._label = None

        self.knowledgepack = model
        self.label = label
        if self.label is None:
            print(
                "Label column is not specified for this knowledgepack, specify as label=XXX"
            )
        self.feature_vector = feature_vector

    @property
    def knowledgepack(self):
        """Auto generated unique identifier for the project object"""
        return self._knowledgepack

    @knowledgepack.setter
    def knowledgepack(self, value):
        if value is None:
            self._knowledgepack = None
        elif hasattr(value, "knowledgepack"):
            self._knowledgepack = value.knowledgepack
        else:
            self._knowledgepack = value

    @property
    def label(self):
        """Auto generated unique identifier for the project object"""
        return self._label

    @label.setter
    def label(self, value):
        if value is None and self.knowledgepack:
            self._label = self.knowledgepack.query_summary.get("label_column", None)
        else:
            self._label = value

    @property
    def feature_vector(self):
        """Auto generated unique identifier for the project object"""
        return self._feature_vector

    @feature_vector.setter
    def feature_vector(self, value):
        self._feature_vector = value[
            [i for i in sorted(value.columns.tolist()) if ("gen_" == i[:4])]
            + [self.label]
        ]

    def plot_features(
        self,
        features_list=None,
        plot_style="boxplot",
        figsize=(20, 5),
        features_per_row=4,
    ):
        """
        Display the distributions of feature values across classes.

        Args:
            features (Dataframe): Feature set that is results of generator_set or feature_selector
            label (String): Name of class label
            features_list (List):  List of features to plot, if None plots all (default: None)
            plot_style (String):  'violinplot' / 'boxplot' to determine style of plot
            figsize (tuple): Size of figure to make
            features_per_row (int): Number of features to show in each row of the plot
        """

        # if features_list is not given use all featues
        if not (features_list):
            features_list = [
                i for i in self.feature_vector.columns.tolist() if ("gen_" == i[:4])
            ]

        # reframe the feature dataframe so it can be used to input for violinplot or boxplot
        num_plots = len(features_list) // (features_per_row)
        if len(features_list) % features_per_row != 0:
            num_plots += 1
        figsize = (figsize[0], figsize[1] * num_plots)
        f, axarr = plt.subplots(num_plots, figsize=figsize)
        set_style("ticks")
        if not isinstance(axarr, np.ndarray):
            axarr = [axarr]

        for index in range(num_plots):
            df_list = [
                reframe(self.feature_vector, feature, self.label)
                for feature in features_list[
                    index * features_per_row : (index + 1) * features_per_row
                ]
            ]

            if plot_style == "violinplot":
                violinplot(
                    x="Features",
                    y="Sensor Values",
                    hue=self.label,
                    data=concat(df_list),
                    palette="muted",
                    ax=axarr[index],
                )
            elif plot_style == "boxplot":
                boxplot(
                    x="Features",
                    y="Sensor Values",
                    hue=self.label,
                    data=concat(df_list),
                    palette="muted",
                    ax=axarr[index],
                )
            else:
                print("Error: plot_style is not in the library")
                return

            axarr[index].set_xticklabels(
                features_list[
                    index * features_per_row : (index + 1) * features_per_row
                ],
                fontsize=12,
                rotation=25,
            )
            axarr[index].set_ylabel("Feature Values", fontsize=16)
            axarr[index].set_xlabel("")
            axarr[index].legend()
            axarr[index].set_ylim((0, 255))
            axarr[index].legend().set_visible(False)

        axarr[-1].set_xlabel("Features", fontsize=16)
        axarr[0].legend(loc="center left", bbox_to_anchor=(1, 0.85), fontsize=16)
        axarr[0].set_title("Feature Vector Distribution", fontsize=16)

        plt.subplots_adjust(hspace=0.9)

        plt.show()

    def neuron_feature_map(
        self,
        feature1,
        features_list,
        label=None,
        norm=None,
        hist_figure=True,
        neuron_figure=True,
        vector_labels=["all"],
        vector_alpha=0.3,
        neuron_alpha=0.2,
        color_list=[
            "m",
            "k",
            "b",
            "lime",
            "r",
            "c",
            "y",
            "brown",
            "tan",
            "g",
            "pink",
            "gray",
        ],
    ):
        """
        This function display neuron map, feature map and feature distributions.

        Args:
            hist_figure = Bool, display historgram of vector for given features.
            neuron_figure = True,
            pivot_feature = String, it is y-axis feature that used to display neuron map, feature map and feature distributions
            versus_features = List<String>, x-axis features that used to display neuron map, feature map and feature distributions
            vector_labels (List): which features will be displayed  [ f1, f2, f3, ..., 'all', [] ], ['all'] display all features, [] does not display any features
            vector_alpha = 0.3, level of vector transparency
            neuron_alpha = 0.1, level of neuron transparency
            norm(int): takes from knowledgepack, to overide 0=L1, 1=LSUP

        Returns: None

        """

        self.hist_figure = hist_figure
        self.pivot_feature = feature1
        self.vector_labels = vector_labels
        self.vector_alpha = vector_alpha
        self.neuron_alpha = neuron_alpha
        self.axis = 10  # axis
        self.color_list = color_list

        if isinstance(features_list, str):
            features_list = [features_list]

        self.columns_index = [None, None]
        self.columns_index[1] = self.feature_vector.columns.tolist().index(
            self.pivot_feature
        )

        for feature in features_list:
            self.columns_index[0] = self.feature_vector.columns.tolist().index(feature)

            if self.vector_labels == ["all"]:
                self.vector_labels = np.unique(self.feature_vector[self.label]).tolist()

            self._set_initial_figure()

            if neuron_figure and self.knowledgepack:
                if norm is None:
                    norm = self.knowledgepack.device_configuration.get(
                        "distance_mode", 0
                    )
                else:
                    norm = norm
                self._plot_neurons(self.knowledgepack.neuron_array, feature, norm)

            if (vector_labels != ["None"]) | (vector_labels == []):
                self._plot_feature_vector(feature)

            self._set_final_figure_settings(feature)
            plt.show()

    def _plot_neurons(self, neurons, versus_feature, norm):
        num = 0
        for neuron in neurons:
            x_val = neuron["Vector"][self.columns_index[0]]
            y_val = neuron["Vector"][self.columns_index[1]]
            aif = neuron["AIF"]

            if neuron["Category"] < 30000:  # if it is not degenerated neuron
                cat_label = self.knowledgepack.class_map[str(neuron["Category"])]
                clr = self.color_list[self.label_indx_list.index(cat_label)]
                if norm == 0:
                    dim = len(neuron["Vector"])
                    new_aif = aif * np.sqrt(2) / dim
                    self.neuron_scatter.add_patch(
                        patches.Rectangle(
                            (x_val - (aif * 2 / dim), y_val),
                            2 * new_aif,
                            2 * new_aif,
                            color=clr,
                            alpha=self.neuron_alpha,
                            angle=-45.0,
                        )
                    )

                else:
                    self.neuron_scatter.add_patch(
                        patches.Rectangle(
                            (x_val - aif, y_val - aif),
                            2 * aif,
                            2 * aif,
                            color=clr,
                            alpha=self.neuron_alpha,
                            angle=0,
                        )
                    )

                self.neuron_scatter.scatter(x_val, y_val, c=clr, marker="*", s=96)

    def _plot_hist(self, Xs, Ys, clr):
        binwidth = 5
        bins = np.arange(0, 256, binwidth)
        self.axHistx.hist(Xs, bins=bins, color=clr, alpha=self.vector_alpha)
        self.axHistx.set_xlim(self.neuron_scatter.get_xlim())

        self.axHisty.hist(
            Ys, bins=bins, color=clr, alpha=self.vector_alpha, orientation="horizontal"
        )
        self.axHisty.set_ylim(self.neuron_scatter.get_ylim())

    def _plot_feature_vector(self, versus_feature):
        for label_indx in self.vector_labels:
            temp_label_data = self.feature_vector[
                self.feature_vector[self.label] == label_indx
            ].index.tolist()
            temp = self.feature_vector.loc[temp_label_data[0]].tolist()
            cat_label = temp[-1]
            clr = self.color_list[self.label_indx_list.index(cat_label)]
            Xs = self.feature_vector.loc[temp_label_data, versus_feature].tolist()
            Ys = self.feature_vector.loc[temp_label_data, self.pivot_feature].tolist()

            self.neuron_scatter.scatter(Xs, Ys, c=clr, alpha=self.vector_alpha)

            if self.hist_figure:
                self._plot_hist(Xs, Ys, clr)

    def _set_initial_figure(self):
        plt.figure(1, figsize=(self.axis, self.axis))

        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        rect_scatter = [left, bottom, width, height]
        self.neuron_scatter = plt.axes(rect_scatter, aspect="equal")

        if self.hist_figure:
            # definitions for the axes
            bottom_h = left_h = left + width + 0.02
            rect_histx = [left, bottom_h, width, 0.2]
            rect_histy = [left_h, bottom, 0.2, height]
            nullfmt = NullFormatter()
            self.axHistx = plt.axes(rect_histx)
            self.axHistx.xaxis.set_major_formatter(nullfmt)
            self.axHisty = plt.axes(rect_histy)
            self.axHisty.yaxis.set_major_formatter(nullfmt)

        if self.knowledgepack:
            self.label_indx_list = np.unique(
                [
                    self.knowledgepack.class_map[str(i)]
                    for i in np.sort(
                        list(map(int, list(self.knowledgepack.class_map.keys())))
                    )
                ]
                + np.unique(self.feature_vector[self.label]).tolist()
            ).tolist()
        else:
            self.label_indx_list = list(np.unique(self.feature_vector[self.label]))

        # this part creates dummy vectors with the unique number of labels for legent.
        for indx in range(len(self.label_indx_list)):
            self.neuron_scatter.scatter(
                0, 0, c=self.color_list[indx], label=self.label_indx_list[indx]
            )

        self.neuron_scatter.set_xlim([0, 256])
        self.neuron_scatter.set_ylim([0, 256])

    def _set_final_figure_settings(self, versus_feature):

        self.neuron_scatter.set_xlabel(versus_feature, fontsize=12)
        self.neuron_scatter.set_ylabel(self.pivot_feature, fontsize=12)

        self.neuron_scatter.set_xticks(range(0, 256, 10))
        self.neuron_scatter.set_xticklabels(range(0, 256, 10), rotation=0, fontsize=8)

        self.neuron_scatter.set_yticks(range(0, 256, 10))
        self.neuron_scatter.set_yticklabels(range(0, 256, 10), rotation=0, fontsize=8)

        if self.hist_figure:
            self.axHistx.set_xticks(range(0, 256, 10))
            self.axHistx.yaxis.set_tick_params(labelsize=8)
            self.axHistx.set_title(self.pivot_feature + " vs " + versus_feature)

            self.axHisty.set_yticks(range(0, 256, 10))
            self.axHisty.xaxis.set_tick_params(labelsize=8)
            self.neuron_scatter.legend(
                loc="center left", bbox_to_anchor=(1.02, 1.26), fontsize=12
            )

        else:
            self.neuron_scatter.legend(
                loc="center left", bbox_to_anchor=(1, 0.945), fontsize=12
            )
            self.neuron_scatter.set_title(self.pivot_feature + " vs " + versus_feature)


def reframe(features, feature, label):
    # This funciton reframes the original features dataframe
    temp_list = DataFrame(columns=[label, "Features", "Sensor Values"])
    temp_list[label] = features[label].tolist()
    temp_list["Features"] = [feature] * len(features)
    temp_list["Sensor Values"] = features[feature].tolist()
    return temp_list



from sensiml.method_calls.basemethodcall import BaseMethodCall


class AugmentationCall(BaseMethodCall):
    """The base class for a augmentation call.

    Child classes have their own additional properties and overwrite this docstring."""

    def __init__(self, name, function_type="Augmentation"):
        super(AugmentationCall, self).__init__(name=name, function_type=function_type)

    def _to_dict(self):
        prop_dict = {}
        prop_dict["inputs"] = {}
        for prop in self._public_properties():
            prop_dict["inputs"][prop] = getattr(self, prop)
        prop_dict["function_name"] = self._name
        return prop_dict



from sensiml.datamanager.pipeline import PipelineStep


class FeatureFileCall(PipelineStep):
    """The base class for a featurefile call"""

    def __init__(self, name):
        super(FeatureFileCall, self).__init__(name=name, step_type="FeatureFileCall")
        self._featurefile = None
        self._feature_columns = None
        self._group_columns = None
        self._label_column = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def feature_columns(self):
        return self._feature_columns

    @feature_columns.setter
    def feature_columns(self, value):
        self._feature_columns = value

    @property
    def group_columns(self):
        return self._group_columns

    @group_columns.setter
    def group_columns(self, value):
        if isinstance(value, list):
            self._group_columns = value
        else:
            print("Group columns must be a list of strings.")

    @property
    def label_column(self):
        return self._label_column

    @label_column.setter
    def label_column(self, value):
        if isinstance(value, str):
            self._label_column = value
        else:
            print("Label Column must be a string")

    @property
    def featurefile(self):
        return self._featurefile

    @featurefile.setter
    def featurefile(self, featurefile):
        self._featurefile = featurefile

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    def _to_dict(self):
        featurefile_dict = {}
        featurefile_dict["name"] = self.name
        featurefile_dict["type"] = "featurefile"
        featurefile_dict["feature_columns"] = self.feature_columns
        featurefile_dict["group_columns"] = self.group_columns
        featurefile_dict["label_column"] = self.label_column
        featurefile_dict["outputs"] = self.outputs
        return featurefile_dict



import sensiml.base.utility as utility
from pandas import DataFrame


class Base(object):
    """
    Base class for all datamanger objects.

    _data (dict): stores the data response from the server
    _fields (list): list of fields that the model has
    _read_only_fields (list): list of fields that are read only
    _field_map (dict): as some fields are named differently from the server
        we use a field map dict for backwards compantibility the field map
        is a dictionary with the following format
        {'client_field_name':'server_field_name'}

    """

    _data = None
    _fields = []
    _read_only_fields = ["uuid", "created_at", "last_modified"]
    _field_map = {}

    def __init__(self, connection, **kwrags):
        """
        All objects take a connection first, followed by some kwargs
        """
        self._connection = connection

    @property
    def base_url(self):
        """Overwrite this property in the subclass for its base url"""

        return ""

    @property
    def detail_url(self):
        """Overwrite this property in the subclass for its base url"""

        return ""

    @property
    def fields(self):
        return self._fields

    @property
    def data(self):
        """
        Returns the server response data object or
        generates the data object from locally stored
        properties
        """
        if self._data is None:
            return [
                getattr(self, field)
                for field in self.fields.keys()
                if hasattr(self, field)
            ]

        return self._data

    def _to_representation(self):
        """
        converts the object into a representation for insert/update
        rest calls
        """
        return {
            self._field_map.get(field, field): getattr(self, field)
            for field in self.fields
            if hasattr(self, field) and field not in self._read_only_fields
        }

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single object.
        stores the results in _data as well

         Args:
             data (dict): server response data object
        """
        for field in self._fields:
            if hasattr(self, field):
                mapped_field = self._field_map.get(field, field)
                setattr(self, field, data.get(mapped_field, None))

        self._data = data

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""

        response = self._connection.request("get", self.detail_url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def delete(self):
        """Calls the REST API and populates the local object properties from the server."""

        response = self._connection.request("delete", self.detail_url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def insert(self, path=None):
        """Calls the REST API to insert a new object."""

        data = self._to_representation()

        if path:
            response = self._connection.file_request(self.base_url, path, data, "rb")
        else:
            response = self._connection.request("post", self.base_url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""

        data = self._to_representation()

        response = self._connection.request("put", self.detail_url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response


class BaseSet(object):
    def __init__(self, connection, initialize_set=True, **kwargs):
        """Initialize a set object to store base objects

        _set (list): a list of objects that are part of the set
        _objclass (Class): the class obj stored in the set
        _attr_key (str): a key used to build out the to_dict

        Args:
            connection (connection) connection object to server
            initialze_set (bool) Default is True. If true will build the
             set of objects.
        """

        self._connection = connection
        self._set = None
        self._objclass = Base
        self._attr_key = "uuid"

        if initialize_set:
            self.refresh()

    @property
    def get_set_url(self):
        """
        replace this with the url to call to pull down the set objects
        """
        pass

    @property
    def objs(self):
        if self._set is None:
            self._set = self.get_set()

        return self._set

    def append(self, obj):
        if self._set is None:
            self._set = [obj]

        self._set.append(obj)

    def to_dict(self, key=None):
        if key is None:
            key = self._attr_key

        return {getattr(k, key): k for k in self.objs}

    def to_dataframe(self):
        return DataFrame([obj.data for obj in self.objs])

    def refresh(self):
        self._set = self.get_set()

    def get_set(self):
        """Calls the REST API to get the set of objects from the server."""
        url = self.get_set_url

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err:
            raise Exception(err)

        # Populate each label from the server
        objs = []
        for obj in response_data:
            objs.append(self._new_obj_from_dict(obj))

        return objs

    def _new_obj_from_dict(self, data):
        """Creates a new object from the response data from the server.

        Args:
            data (dict): contains properties of the object

        Returns:
            obj of type _objclass

        """
        obj = self._objclass(self._connection, project=self._project)
        obj.initialize_from_dict(data)
        return obj



import datetime
import os

import sensiml.base.utility as utility
from sensiml.datamanager.base import Base


class FeatureFile(Base):
    """Base class for a featurefile object."""

    _fields = [
        "uuid",
        "name",
        "created_at",
        "last_modified",
        "is_features",
        "label_column",
        "number_rows",
    ]

    def __init__(
        self,
        connection,
        project,
        name="",
        path="",
        is_features=True,
        uuid=None,
        label_column="",
        number_rows=None,
    ):
        self._connection = connection
        self._project = project
        self.uuid = uuid
        self._created_at = None
        self._is_features = is_features
        self.name = name
        self.path = path
        self.label_column = label_column
        self.number_rows = number_rows

    # Maintain campatibility with old filename attr
    @property
    def filename(self):
        """The name of the file as stored on the server

        Note:
            Filename must contain a .csv or .arff extension
        """
        return self.name

    @filename.setter
    def filename(self, value):
        self.name = value

    @property
    def is_features(self):
        """If this is a DataFile or FeatureFile"""
        return self._is_features

    @is_features.setter
    def is_features(self, value):
        self._is_features = value

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        self._created_at = datetime.datetime.strptime(
            value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
        )

    def insert(self):
        """Calls the REST API to insert a new featurefile."""
        url = "project/{0}/featurefile/".format(self._project.uuid)
        featurefile_info = {"name": self.name, "is_features": self.is_features}

        if not os.path.exists(self.path):
            raise OSError(
                "Cannot update featurefile because the system cannot find the file path: '{0}'. "
                "Please update the file's path attribute.".format(self.path)
            )

        response = self._connection.file_request(url, self.path, featurefile_info, "rb")
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

        return response

    def update(self):
        """Calls the REST API to update the featurefile's properties on the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        featurefile_info = {"name": self.name, "is_features": self.is_features}

        if not os.path.exists(self.path):
            raise OSError(
                "Cannot update featurefile because the system cannot find the file path: '{0}'. "
                "Please update the file's path attribute.".format(self.path)
            )

        response = self._connection.file_request(
            url, self.path, featurefile_info, "rb", method="put"
        )
        response_data, err = utility.check_server_response(response)

        return response

    def delete(self):
        """Calls the REST API and deletes the featurefile from the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populate the featurefile's properties from the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data["name"]
            self.is_features = response_data["is_features"]

        return response

    def download(self):
        """Calls the REST API and retrieves the featurefile's binary data.

        Returns:
            featurefile contents
        """

        url = "project/{0}/featurefile/{1}/data/".format(self._project.uuid, self.uuid)

        response = self._connection.request("get", url)

        return response

    def download_json(self):
        """Calls the REST API and retrieves the featurefile's json data.

        Returns:
            featurefile contents as json
        """

        url = "project/{0}/featurefile/{1}/json/".format(self._project.uuid, self.uuid)

        response = self._connection.request("get", url)

        return response

    def compute_analysis(self, analysis_type="UMAP", **kwargs):
        """Calls the REST API to compute the analysis for the feature file.

        Args:
            analysis_type (str): the type of clustering analysis, ie "UMAP" (default), "TSNE" and "PCA".

        Kwargs:
            shuffle_seed (int): random seed to shuffle and resample feature vector
            analysis_seed (int): random state of the analysis (default is 0)
            n_neighbor (int): The size of local neighborhood (in terms of number of neighboring sample points) used for manifold approximation. If not specified, default is the number of unique labels.
            n_components (int): The dimension of the output result. Default is 2. 'n_components' is adjusted based on the method, dimension of the feature vector and number of samples
            n_sample (int): Maximum number of output samples. Default is 1000.

        Returns:
            A JSON response containing the metadata of the generated analysis.

        Example:
            >>> feature_file = client.get_featurefile(<feature-file uuid>)
            >>> response = feature_file.compute_analysis(analysis_type="PCA", shuffle_seed=13, n_components=5)
            >>> response.json()

        """

        url = "project/{0}/featurefile-analysis/{1}/".format(
            self._project.uuid, self.uuid
        )

        json_payload = dict({"analysis_type": analysis_type}, **kwargs)

        response = self._connection.request("post", url, json=json_payload)

        return response

    def list_analysis(self):
        """Calls the REST API and retrieve list of computed analysis for featurefile

        Returns:
            JSON response holding the list of all computed analysis
        """

        url = "project/{0}/featurefile-analysis/{1}/".format(
            self._project.uuid, self.uuid
        )

        response = self._connection.request("get", url)

        response_data, err = utility.check_server_response(response)

        if err is True:
            print(err)
            return response

        return response_data



import copy
import itertools

from pandas import DataFrame
from tabulate import tabulate


class ConfusionMatrix(object):
    """A class used to create, store, and display a confusion matrix based on predicted segments and ground truth segments and compare against other confusion matrices
    Attributes:
        prediction_dict (dict): A dictionary of dictionaries that contains the confusion matrix prediction counts
        matrix (list): A formatted list of lists containing the confusion matrix data to be displayed
        model (str, optional): Name of model used to make prediction files (model being tested). Defaults to "".
        model_session (str, optional): Name of session that predcition files are labeled in. Defaults to "".
        gt_session (str, optional): Name of session that ground truth files are labeled in. Defaults to "".
        overlap_percent (float, optional): Percentage of overlap between predicted segment and real segment in order for the prediction to be considered over that even. Defaults to 50.0.
    Methods:
        compare_confusion_matrices(confusion_matrix_b)
            Compare this matrix to another and return matrix where every value is the difference in the values at the same [i][j] index in matrix a and matrix b
        display(matrix_to_display=None)
            Prints a tabulated ConfusionMatrix (prints self.matrix if no matrix is entered)
    """

    __BLACK = "\033[1;30m"
    __RED = "\033[1;31m"
    __GREEN = "\033[1;32m"
    __YELLOW = "\033[1;33m"
    prediction_dict = {}
    matrix = []
    model_session = ""
    gt_session = ""
    model = ""
    __predicted_segments = []
    __ground_truth_segments = []
    overlap_percent = 50.0

    def __init__(
        self,
        pred_segments: list,
        gt_segments: list,
        model_name: str = "",
        pred_session_name: str = "",
        gt_session_name: str = "",
        overlap_percent: float = 50.0,
    ):
        """Creates a Confusion Matrix object based on the given predicted segments and ground truth segments
        Args:
            pred_segments (list): list of (or single) datasegments or dataframe objects from prediction files to create confusion matrix from
            gt_segments (list): list of (or single) datasegments or dataframe objects from ground truth files to create confusion matrix from
            model_name (str, optional): Name of model used to make prediction files (model being tested). Defaults to "".
            pred_session_name (str, optional): Name of session that predcition files are labeled in. Defaults to "".
            gt_session_name (str, optional): Name of session that ground truth files are labeled in. Defaults to "".
            overlap_percent (float, optional): Percentage of overlap between predicted segment and real segment in order for the prediction to be considered over that even. Defaults to 50.0.
        """
        self.overlap_percent = overlap_percent
        self.model_session = pred_session_name
        self.gt_session = gt_session_name
        self.model = model_name
        pred_segments = copy.deepcopy(pred_segments)
        gt_segments = copy.deepcopy(gt_segments)
        if not isinstance(pred_segments, list):
            pred_segments = [pred_segments]
        if not isinstance(gt_segments, list):
            gt_segments = [gt_segments]
        for i in range(0, len(pred_segments)):
            if not isinstance(pred_segments[i], DataFrame):
                pred_segments[i] = pred_segments[i].to_dataframe()
        for i in range(0, len(gt_segments)):
            if not isinstance(gt_segments[i], DataFrame):
                gt_segments[i] = gt_segments[i].to_dataframe()

        self.matrix = self.__compare_predictions(pred_segments, gt_segments)
        self.__ground_truth_segments = gt_segments
        self.__predicted_segments = pred_segments

    def __repr__(self):
        self.display()
        return ""

    def __str__(self):
        self.display()
        return ""

    def __get_segment_list(self, segments_df: DataFrame) -> list:
        """take segments dataframe and returns a list of dictionaries, where each dictionary belongs to a different segment, with key names value, start, and end
        Args:
            segments_df (DataFrame): dataframe of segments converted from DataSegments object
        Returns:
            list: list of dictionaries containing condensed segment information
        """
        segments_df = segments_df.rename(
            columns={
                "label_value": "value",
                "capture_sample_sequence_start": "start",
                "capture_sample_sequence_end": "end",
            }
        )
        segment_list = segments_df.to_dict("records")
        return segment_list

    def __get_percent(self, numerator: int, denominator: int) -> float:
        """calculate percents rounded to the hundredths place while accounting for 0/0
        Args:
            numerator (int): numerator of the fraction corresponding to the percentage
            denominator (int): denominator of the fraction corresponding to the percentage
        Returns:
            float: the percent rounded to the hundredths place corresponding to (100*numerator/denominator) (or 0 in the case of 0/0)
        """
        return (
            round(100 * numerator / denominator, 2) if denominator and numerator else 0
        )

    def __calculate_overlap_percent(
        self, start_1: int, start_2: int, end_1: int, end_2: int
    ) -> float:
        """calculate how much of label_2 [start_2, end_2] is covered by label_1 [start_1, end_1] as a percent
        using the formula: 100*(min(end_1, end_2) - max(start_1, start_2))/(length_label_2)
        Args:
            start_1 (int): start value of label 1 (usually the prediction label)
            start_2 (int): start value of label 2 (usually the real label)
            end_1 (int): end value of label 1 (usually the prediction label)
            end_2 (int): start value of label 2 (usually the real label)
        Returns:
            float: the percent of label_2 that label_1 covers
        """
        # calculate overlap between labels
        label_overlap = min(end_1, end_2) - max(start_1, start_2)
        # calculate length of label 2
        range_of_label_2 = end_2 - start_2
        # return overlap percent
        return self.__get_percent(label_overlap, range_of_label_2)

    def __get_rc_list(
        self,
        pred_segment_list: list,
        real_segment_list: list,
        is_col_list: bool = False,
    ) -> list:
        """build the list of all labels found in the segment lists and add the names of the totals and % true positives
        column/row names based on whether a column name list or row name list is asked for
        Args:
            pred_segment_list (list): list of segments predicted by model
            real_segment_list (list): list of ground truth/ manually labeled segments
            is_col_list (bool, optional): boolean indicating whether to return a list of column names or row names. Defaults to False.
        Returns:
            list: list of the row or column names based on the segment labels in the predicted and real segment lists
        """

        events = [" "]
        for all_segment_set in list(
            itertools.zip_longest(pred_segment_list, real_segment_list)
        ):
            segment_set_all = [x for x in all_segment_set if not x == None]
            for segment_set in segment_set_all:
                for segment in segment_set:
                    label = segment["value"]
                    if events.count(label) == 0:
                        events.append(label)
        events.sort()
        events.append("UNK")
        if not is_col_list:
            events.extend(["Total", "Pos_Pred(%)"])
        else:
            events.extend(["Support", "Sense %"])
        return events

    def __build_matrix_comparison_frame(self, conf_matrix: list) -> list:
        """build a list of lists that copies column and row names from the inputed matrix
        Args:
            conf_matrix (list): confusion matrix to copy row and col names from
        Returns:
            list: list of lists where the 1st row are the column names and each successive row starts with the proper row names (no inner values)
        """
        tmp = [conf_matrix[0]]
        for name in tmp[0]:
            name = ConfusionMatrix.__BLACK + name
        for i in range(1, len(conf_matrix)):
            tmp.append([ConfusionMatrix.__BLACK + conf_matrix[i][0]])
        return tmp

    def __get_default_matrix_alignment(self, num_cols: int) -> list:
        """return list of "default" alignment, where the first column is left justified and all other columns are right justified
        Args:
            num_cols (int): number of columns in matrix
        Returns:
            list: list of "default" alignment, where the first column is left justified and all other columns are right justified
        """
        align_list = ["left"]
        for i in range(1, num_cols):
            align_list.append("right")
        return align_list

    def __insert_horizontal_line_table(
        self, table: list, pos: int, alignment: list
    ) -> list:

        """add a horizontal breaker line into a table
        Args:
            table (list): table in the form of a list of lists
            pos (int): row index for the line to be inserted at
            alignment (list): list of column alignments for tabulate format
        Returns:
            list: list of lists identical to the inputed list of lists with an inputed horizontal line row at index "pos"
        """
        tabulated = tabulate(table, colalign=alignment)
        newline_ind = tabulated.index("\n")
        line_breaker = tabulated[0:newline_ind]
        breaker = line_breaker.split()
        table.insert(pos, breaker)
        return table

    def __get_sum_row(self, pred_cts: dict, event: str, col_names: list) -> int:
        """calculate the sum of the numerical elements in a given row
        Args:
            pred_cts (dict): dictionary containing all predictions made and how they map to the actual labels
            event (str): name of the event row that you want to find the sum total of
            col_names (list): list of the column names to iterate over
        Returns:
            int: sum total of items in the event row
        """
        sum = 0
        for x in range(1, len(col_names) - 2):
            sum += pred_cts[col_names[x]][event]
        return sum

    def __get_sum_col(self, pred_cts: dict, event: str, row_names: list) -> int:
        """calculate the sum of the numerical elements in a given column
        Args:
            pred_cts (dict): dictionary containing all predictions made and how they map to the actual labels
            event (str): name of the event column that you want to find the sum total of
            row_names (list): list of the row names to iterate over
        Returns:
            int: sum total of items in the event column
        """
        sum = 0
        for x in range(1, len(row_names) - 2):
            sum += pred_cts[event][row_names[x]]
        return sum

    def __build_pred_dict(self, row_names: list, col_names: list) -> dict:
        """build a dictionary of dictionaries using the row names as the outer key set and the column names as the keys in each inner dictionary
        Args:
            row_names (list): list of the row names
            col_names (list): list of the column names
        Returns:
            dict: dictionary of dictionaries using the row names as the outer key set and the column names as the keys in each inner dictionary
        """
        tmp_dict = {}
        for i in range(1, len(row_names) - 2):
            tmp_dict[row_names[i]] = {}
            for j in range(1, len(col_names) - 2):
                tmp_dict[row_names[i]][col_names[j]] = 0
        return tmp_dict

    def __build_event_dict(self, row_names: list) -> dict:
        """build dictionary with keys as the different events (given in the row names list)
        Args:
            row_names (list): list of the row names
        Returns:
            dict: dictionary with keys as the different events (given in the row names list)
        """
        tmp_dict = {}
        for i in range(1, len(row_names) - 2):
            tmp_dict[row_names[i]] = 0
        return tmp_dict

    def __fill_sum_dict(
        self, pred_cts: dict, row_names: list, col_names: list, sum_type: str
    ) -> dict:
        """fill a dictionary with "row sums" or "column sums" (totals & supports)
        Args:
            pred_cts (dict): dictionary containing all predictions made and how they map to the actual labels
            row_names (list): list of the row names
            col_names (list): list of the column names
            sum_type (str): string indicating whether the dictionary should contain the row sums (support) or column sums (total)
        Returns:
            dict: dictionary containing "row sums" or "column sums" (totals & supports) where the keys are the row/col names
        """
        event_dict = self.__build_event_dict(row_names)
        for i in range(1, len(row_names) - 2):
            if sum_type == "r":
                event_dict[row_names[i]] = self.__get_sum_row(
                    pred_cts, row_names[i], col_names
                )
            else:
                event_dict[row_names[i]] = self.__get_sum_col(
                    pred_cts, row_names[i], row_names
                )
        return event_dict

    def __fill_percent_dict(
        self,
        pred_cts: dict,
        row_sums: dict,
        col_sums: dict,
        row_names: list,
        sum_type: str,
    ) -> dict:
        """fill a dictionary with "sense%" or "pos pred %"
        Args:
            pred_cts (dict): dictionary containing all predictions made and how they map to the actual labels
            row_sums (dict): list of the row sums
            col_sums (dict): list of the column sums
            row_names (list): list of the row names
            sum_type (str): list of the column names
        Returns:
        dict : dictionary containing the "sense%" or "pos pred %" where the keys are the row/col names
        """
        perc_dict = self.__build_event_dict(row_names)
        for i in range(1, len(row_names) - 2):
            if sum_type == "s":
                perc_dict[row_names[i]] = self.__get_percent(
                    pred_cts[row_names[i]][row_names[i]], row_sums[row_names[i]]
                )
            else:
                perc_dict[row_names[i]] = self.__get_percent(
                    pred_cts[row_names[i]][row_names[i]], col_sums[row_names[i]]
                )
        return perc_dict

    def __fill_prediction_dict(
        self, pred_segments: list, ground_truth_segments: list, pred_cts: dict
    ) -> dict:
        """fill a dictionary of dictionaries where the outer keys are the actual events and the inner keys are
        what they are predicted as (all labels that overlap with less than 40% of a real label are counted as UNK)
        Args:
            pred_segments (list): list of segments from prediction file
            ground_truth_segments (list): list of segments from manually labeled file
            pred_cts (dict): Empty dictionary of dictionaries with event keys
        Returns:
            dict: dictionary of dictionaries where the outer keys are the actual events and the inner keys are what they
                are predicted as with the inner values are the counts of each label
        """
        # Loop over all predicted segments
        for pdictionary in pred_segments:
            # Loop over real segments to find all real segements that the predicted segment overlaps with
            # (discard predictions that overlap w/ nothing)
            num_overlaps = 0
            for adictionary in ground_truth_segments:
                if len(adictionary) == 0:
                    break
                # If the next real segment begins after the predicted one ends then there will be no more overlapping segments
                if adictionary["start"] > pdictionary["end"]:
                    break
                # If the next real segment ends before the predicted one starts then it won't overlap so we should skip it
                if adictionary["end"] < pdictionary["start"]:
                    continue
                overlap = self.__calculate_overlap_percent(
                    pdictionary["start"],
                    adictionary["start"],
                    pdictionary["end"],
                    adictionary["end"],
                )
                # Keep track of what the prediction was and how it compared to the actual label IF it is over the overlap threshold
                if overlap >= self.overlap_percent:
                    pred_cts[pdictionary["value"]][adictionary["value"]] += 1
                    num_overlaps += 1
            if num_overlaps == 0:
                pred_cts[pdictionary["value"]]["UNK"] += 1

        # Count number of real segments that were left unlabeled in predictions
        for adictionary in ground_truth_segments:
            num_overlaps = 0
            for pdictionary in pred_segments:
                if pdictionary["start"] > adictionary["end"]:
                    break
                if pdictionary["end"] < adictionary["start"]:
                    continue
                num_overlaps += 1
            if num_overlaps == 0:
                pred_cts["UNK"][adictionary["value"]] += 1
        return pred_cts

    # Method to build the confusion matrix
    def __build_conf_matrix(
        self,
        pred_cts: dict,
        row_names: list,
        col_names: list,
        row_sum_dict: dict,
        col_sum_dict: dict,
        sense_dict: dict,
        pospred_dict: dict,
        sum_total: int,
    ) -> list:

        """list of lists containing the confusion matrix data
        Args:
            pred_cts (dict): Dictionary containing the segment prediction data
            row_names (list): List of the confusion matrix row names
            col_names (list): List of the confusion matrix col names
            row_sum_dict (dict): Dictionary containing the sums of each row
            col_sum_dict (dict): Dictionary containing the sums of each col
            sense_dict (dict): Dictionary containing the sensitivity for each row
            pospred_dict (dict): Dictionary containing the positive prediction % for each col
            sum_total (int): Sum total of all predictions made
        Returns:
            list: build and format confusion matrix
        """
        conf_matrix = []
        conf_matrix.append(col_names)
        corr_sum = 0
        for x in range(1, len(row_names) - 2):
            label = row_names[x]
            l = [label]
            for y in range(1, len(row_names) - 2):
                if x == y:
                    corr_sum += pred_cts[row_names[y]][label]
                    l.append(
                        ConfusionMatrix.__GREEN + str(pred_cts[row_names[y]][label])
                    )
                else:
                    l.append(ConfusionMatrix.__RED + str(pred_cts[row_names[y]][label]))
            l.append(ConfusionMatrix.__BLACK + str(row_sum_dict[label]))
            l.append(ConfusionMatrix.__BLACK + str(sense_dict[label]))
            conf_matrix.append(l)
        l = ["Total"]
        l.extend(list(col_sum_dict.values()))
        l.append(sum_total)
        conf_matrix.append(l)
        l = ["Pos_Pred(%)"]
        l.extend(list(pospred_dict.values()))
        l.extend(["Acc(%)", self.__get_percent(corr_sum, sum_total)])
        conf_matrix.append(l)

        return conf_matrix

    def __compare_predictions(
        self, pred_segments: list, ground_truth_segments: list
    ) -> list:

        """compare segments from predicted file(s) and ground truth file(s) and return a confusion matrix from the results. The number of prediction files and ground truth files MUST match.
        Args:
            pred_segments (list): list of segments from prediction file each set of segments in dataframe format
            ground_truth_segments (list): list of segments from manually labeled file each set of segments in dataframe format
        Returns:
            list: list: a list of lists with the confusion matrix data
        """
        if len(pred_segments) != len(ground_truth_segments):
            raise Exception(
                "The number of prediction files and ground truth files do not currently match."
            )
        for i in range(0, len(pred_segments)):
            pred_segments[i] = self.__get_segment_list(pred_segments[i])
            ground_truth_segments[i] = self.__get_segment_list(ground_truth_segments[i])
        row_names = self.__get_rc_list(
            pred_segments,
            ground_truth_segments,
            is_col_list=False,
        )
        col_names = self.__get_rc_list(
            pred_segments,
            ground_truth_segments,
            is_col_list=True,
        )
        # Count of labels in capture file labeled by model
        pred_cts = self.__build_pred_dict(row_names, col_names)
        for index in range(0, len(pred_segments)):
            pred_cts = self.__fill_prediction_dict(
                pred_segments[index], ground_truth_segments[index], pred_cts
            )
        self.prediction_dict = copy.deepcopy(pred_cts)
        # Get row sums
        row_sum_dict = self.__fill_sum_dict(pred_cts, row_names, col_names, "r")
        # Get col sums
        col_sum_dict = self.__fill_sum_dict(pred_cts, row_names, col_names, "c")
        # Get total sum
        sum_total = sum(row_sum_dict.values())
        # Get sense %s
        sense_dict = self.__fill_percent_dict(
            pred_cts, row_sum_dict, col_sum_dict, row_names, "s"
        )
        # Get positive prediction %s
        pospred_dict = self.__fill_percent_dict(
            pred_cts, row_sum_dict, col_sum_dict, row_names, "p"
        )

        confusion_matrix = self.__build_conf_matrix(
            pred_cts,
            row_names,
            col_names,
            row_sum_dict,
            col_sum_dict,
            sense_dict,
            pospred_dict,
            sum_total,
        )

        return confusion_matrix

    def __print_confusion_matrix(
        self,
        row_names: list = None,
        col_names: list = None,
        colalign: list = None,
    ) -> None:
        """prints a confusion matrix with prediction session name and model name as title using the given segments and alignment
        Args:
            row_names (list, optional): list of row names. Defaults to None.
            col_names (list, optional): list of column names. Defaults to None.
            colalign (list, optional): list of alignments. Defaults to None.
        """
        self.__ground_truth_segments
        self.__predicted_segments
        print(ConfusionMatrix.__BLACK + "Pred Session: " + self.model_session)
        print("Model: " + self.model)
        if row_names == None:
            row_names = self.__get_rc_list(
                self.__predicted_segments,
                self.__ground_truth_segments,
                is_col_list=False,
            )
        if col_names == None:
            col_names = self.__get_rc_list(
                self.__predicted_segments,
                self.__ground_truth_segments,
                is_col_list=True,
            )
        if colalign == None:
            colalign = self.__get_default_matrix_alignment(len(col_names))
        self.matrix = self.__insert_horizontal_line_table(
            self.matrix, len(self.matrix) - 2, colalign
        )
        print(tabulate(self.matrix, colalign=colalign))

    def compare_confusion_matrices(
        self,
        conf_matrix_b: list,
    ) -> list:
        """compare this matrix to another and create matrix where every value is the difference in the values at the same [i][j] index in matrix a and matrix b
        Args:
            conf_matrix_b (list): confusion matrix to set as the theoretically worse matrix (generally entered as other.matrix where other is a different ConfusionMatrix object)
        Returns:
            list: matrix where every value is the difference in the values at the same [i][j] index in matrix a and matrix b
                (values where a[i][j] > b[i][j] are colored GREEN and values where a[i][j] < b[i][j] are colored RED)
        """

        colalign = ["left"]
        for i in range(1, len(self.matrix[0])):
            colalign.append("right")
        tmp = self.__build_matrix_comparison_frame(self.matrix)
        for i in range(1, len(self.matrix)):
            for j in range(1, len(self.matrix[i])):
                a = (
                    str(self.matrix[i][j])
                    .replace(ConfusionMatrix.__BLACK, "")
                    .replace(ConfusionMatrix.__GREEN, "")
                    .replace(ConfusionMatrix.__RED, "")
                )
                b = (
                    str(conf_matrix_b[i][j])
                    .replace(ConfusionMatrix.__BLACK, "")
                    .replace(ConfusionMatrix.__GREEN, "")
                    .replace(ConfusionMatrix.__RED, "")
                )
                if a.replace(".", "").isdecimal() and b.replace(".", "").isdecimal():
                    num = float(a) - float(b)
                    if (num).is_integer():
                        num = int(num)
                    if num != 0 and (
                        i == len(self.matrix) - 2 or j == len(self.matrix) - 2
                    ):
                        tmp[i].append(ConfusionMatrix.__YELLOW + str(round(num, 2)))
                    elif num > 0 and (
                        (i == j)
                        or (i == len(self.matrix) - 1 or j == len(self.matrix) - 1)
                    ):
                        tmp[i].append(
                            ConfusionMatrix.__GREEN + "+" + str(round(num, 2))
                        )
                    elif num < 0 and (
                        i != j
                        and (i != len(self.matrix) - 1 and j != len(self.matrix) - 1)
                    ):
                        tmp[i].append(ConfusionMatrix.__GREEN + str(round(num, 2)))
                    elif num == 0:
                        tmp[i].append(ConfusionMatrix.__BLACK + str(round(num, 2)))
                    elif num > 0:
                        tmp[i].append(ConfusionMatrix.__RED + "+" + str(round(num, 2)))
                    else:
                        tmp[i].append(
                            ConfusionMatrix.__RED + "-" + str(abs(round(num, 2)))
                        )

                else:
                    tmp[i].append(ConfusionMatrix.__BLACK + a)
        return tmp

    def display(self, matrix_to_display=None):
        """Displays a tabulated ConfusionMatrix
        Args:
            matrix_to_display (ConfusionMatrix, optional): _description_. Defaults to None, in which case self will be displayed.
        """
        if matrix_to_display != None:
            num_cols = len(matrix_to_display[0])
            colalign = self.__get_default_matrix_alignment(num_cols)
            print(tabulate(matrix_to_display, colalign=colalign))
        else:
            self.__print_confusion_matrix()



import os
import json
from sensiml.base import utility
from sensiml.datamanager.base import Base, BaseSet


class LibraryPack(Base):
    """Base class for a library pack object"""

    _uuid = ""
    _name = ""
    _build_version = None
    _maintainer = ""
    _description = ""

    _fields = [
        "uuid",
        "name",
        "build_version",
        "description",
        "maintainer",
    ]

    _read_only_fields = ["uuid", "build_version"]

    def __init__(self, connection, uuid=None):
        self._connection = connection
        if uuid:
            self.uuid = uuid
            self.refresh()

    @property
    def base_url(self):
        return "library-pack/"

    @property
    def detail_url(self):
        return "library-pack/{uuid}/".format(uuid=self.uuid)

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def build_version(self):
        print(self._build_version)

    @build_version.setter
    def build_version(self, value):
        self._build_version = value

    @property
    def maintainer(self):
        return self._maintainer

    @maintainer.setter
    def maintainer(self, value):
        self._maintainer = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""

        response = self._connection.request("get", self.detail_url)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def delete(self):
        """Calls the REST API and populates the local object properties from the server."""

        response = self._connection.request("delete", self.detail_url)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def insert(self):
        """Calls the REST API to insert a new object."""

        data = self._to_representation()

        response = self._connection.request("post", self.base_url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""

        data = self._to_representation()

        response = self._connection.request("put", self.detail_url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response


class LibraryPackSet(BaseSet):
    def __init__(self, connection, initialize_set=True):
        """Initialize a libraryPack object.

        Args:
            connection
        """
        self._connection = connection
        self._set = None
        self._objclass = LibraryPack
        self._attr_key = "uuid"

        if initialize_set:
            self.refresh()

    @property
    def library_packs(self):
        return self.objs

    @property
    def get_set_url(self):
        return "library-pack/"

    def _new_obj_from_dict(self, data):
        """Creates a new object from the response data from the server.

        Args:
            data (dict): contains properties of the object

        Returns:
            obj of type _objclass

        """
        obj = self._objclass(self._connection)
        obj.initialize_from_dict(data)
        return obj



import logging

import sensiml.base.utility as utility
from sensiml.datamanager.base import Base, BaseSet

logger = logging.getLogger(__name__)


class MetadataType(object):
    Int = "integer"
    Float = "float"
    String = "string"


class MetadataRelationship(Base):
    """Base class for a metadata object."""

    _fields = ["uuid", "metadata", "metadata_value", "created_at", "last_modified"]
    _field_map = {"metadata": "label", "metadata_value": "label_value"}

    def __init__(
        self, connection, project, capture=None, metadata=None, metadata_value=None
    ):
        """
        Initialize a metadata object.

            Args:
                connection
                project
                capture

        """

        self._connection = connection
        self._uuid = None
        self._project = project
        self._capture = capture
        self._metadata = metadata
        self._metadata_value = metadata_value

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def capture(self):
        if isinstance(self._capture, str):
            return self._capture
        else:
            return self._capture.uuid

    @capture.setter
    def capture(self, value):
        self._capture = value

    @property
    def metadata(self):
        if isinstance(self._metadata, str):
            return self._metadata
        else:
            return self._metadata.uuid

    @metadata.setter
    def metadata(self, value):
        self._metadata = value

    @property
    def metadata_value(self):
        if isinstance(self._metadata_value, str):
            return self._metadata_value
        else:
            return self._metadata_value.uuid

    @metadata_value.setter
    def metadata_value(self, value):
        self._metadata_value = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""

        url = "v2/project/{0}/capture/{1}/metadata-relationship/".format(
            self._project.uuid, self.capture
        )

        data = self._to_representation()

        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "v2/project/{0}/capture/{1}/metadata-relationship/{2}/".format(
            self._project.uuid, self.capture, self.uuid
        )

        data = self._to_representation()

        response = self._connection.request("put", url, data)
        response_data, err = utility.check_server_response(response)

        return response

    def delete(self):
        """Calls the REST API and deletes the object from the server."""

        url = "v2/project/{0}/capture/{1}/metadata-relationship/{2}/".format(
            self._project.uuid, self.capture, self.uuid
        )

        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""

        url = "v2/project/{0}/capture/{1}/metadata-relationship/{2}/".format(
            self._project.uuid, self.capture, self.uuid
        )

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

        Args:
            dict (dict): contains the uuid, name, type, value, properties
        """

        if self._capture is None:
            self._capture = data["capture"]

        super(MetadataRelationship, self).initialize_from_dict(data)


class MetadataRelationshipSet(BaseSet):
    """Base class for a collection of metadata"""

    def __init__(self, connection, project, capture, initialize_set=True):
        self._connection = connection
        self._project = project
        self._capture = capture

        self._set = None
        self._objclass = MetadataRelationship
        self._attr_key = "uuid"

        if initialize_set:
            self.refresh()

    @property
    def capture(self):
        if isinstance(self._capture, str):
            return self._capture
        else:
            return self._capture.uuid

    @property
    def get_set_url(self):
        # Query the server and get the json
        return "v2/project/{0}/capture/{1}/metadata-relationship/".format(
            self._project.uuid, self.capture
        )

    def _new_obj_from_dict(self, data):
        """Creates a new metadata from data in the dictionary.

        Args:
            dict (dict): contains metadata properties uuid, name, type, value,

        Returns:
            metadata object

        """

        metadata = MetadataRelationship(self._connection, self._project, self._capture)
        metadata.initialize_from_dict(data)

        return metadata


class BulkMetadataRelationshipSet(MetadataRelationshipSet):
    """Base class for a collection of metadata"""

    def __init__(self, connection, project, initialize_set=True):
        self._connection = connection
        self._project = project
        self._objclass = MetadataRelationship
        self._attr_key = "uuid"

        if initialize_set:
            self.refresh()

    def bulk_create(self, metadataset=None):
        """
        Calls the REST API and bulk inserts objects onto the server using the local object's properties.
        """

        if metadataset:
            self._set = metadataset

        url = "project/{0}/metadata-relationship/".format(self._project.uuid)

        data = [
            {
                "label": obj.metadata,
                "label_value": obj.metadata_value,
                "capture": obj.capture,
            }
            for obj in self.objs
        ]

        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)

        if err:
            raise Exception(err)

        # Populate each label from the server
        objs = []
        for obj in response_data:
            objs.append(self._new_obj_from_dict(obj))

        self._set.extend(objs)

        return objs

    @property
    def get_set_url(self):
        # Query the server and get the json
        return "project/{0}/metadata-relationship/".format(self._project.uuid)

    def _new_obj_from_dict(self, data):
        """Creates a new metadata from data in the dictionary.

        Args:
            dict (dict): contains metadata properties uuid, name, type, value,

        Returns:
            metadata object

        """
        metadata = MetadataRelationship(self._connection, self._project)
        metadata.initialize_from_dict(data)

        return metadata



from sensiml.datamanager.capture_configurations import CaptureConfigurations
from sensiml.datamanager.clientplatformdescriptions import ClientPlatformDescriptions
from sensiml.datamanager.featurefiles import FeatureFiles
from sensiml.datamanager.functions import Functions
from sensiml.datamanager.projects import Projects
from sensiml.datamanager.segmenter import SegmenterSet
from sensiml.datamanager.team import Team

__all__ = [
    "Functions",
    "FeatureFiles",
    "Projects",
    "SegmenterSet",
    "ClientPlatformDescriptions",
    "CaptureConfigurations",
    "Team",
]



import logging

import sensiml.base.utility as utility
from sensiml.datamanager.base import Base, BaseSet

logger = logging.getLogger(__name__)


class LabelValue(object):
    Int = "integer"
    Float = "float"
    String = "string"


class LabelValue(Base):
    """Base class for a label object."""

    _fields = ["uuid", "value", "created_at", "last_modified", "color"]

    def __init__(self, connection, project, label):
        """Initialize a metadata object.

        Args:
            connection
            project
            label
        """
        if label._metadata:
            raise Exception("Must be label not metadata")

        self._uuid = ""
        self._value = ""
        self._color = None
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project
        self._label = label

    @property
    def uuid(self):
        """Auto generated unique identifier for the  label value object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def value(self):
        """The data type of the label value object"""
        return self._value

    @value.setter
    def value(self, value):
        """The data type of the label value object"""
        self._value = value

    @property
    def created_at(self):
        """The created time of the label value object"""
        return self._created_at

    @property
    def label(self):
        return self._label

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/{1}/{2}/labelvalue/".format(
            self._project.uuid, self._label._label_or_metadata, self._label.uuid
        )
        data = {"value": self.value, "color": self.color}
        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "project/{0}/{1}/{2}/labelvalue/{3}".format(
            self._project.uuid,
            self._label._label_or_metadata,
            self._label.uuid,
            self.uuid,
        )

        data = self._to_representation()

        response = self._connection.request("put", url, data)
        response_data, err = utility.check_server_response(response)

        return response

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/{1}/{2}/labelvalue/{3}".format(
            self._project.uuid,
            self._label._label_or_metadata,
            self._label.uuid,
            self.uuid,
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/{1}/{2}/labelvalue/{3}".format(
            self._project.uuid,
            self._label._label_or_metadata,
            self._label.uuid,
            self.uuid,
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def initialize_from_dict(self, data):
        """Reads a json dictionary and populates a single metadata object.

        Args:
            dict (dict): contains the uuid, value
        """
        self.uuid = data["uuid"]
        self.value = data["value"]
        self.color = data["color"]
        self._data = data


class LabelValueSet(BaseSet):
    def __init__(self, connection, project, label, initialize_set=True):
        """Initialize a metadata object.

        Args:
            connection
            project
        """
        self._connection = connection
        self._project = project
        self._label = label
        self._set = None
        self._objclass = LabelValue
        self._attr_key = "value"

        if initialize_set:
            self.refresh()

    @property
    def label_values(self):
        return self.objs

    @property
    def get_set_url(self):
        return "project/{0}/{1}/{2}/labelvalue/".format(
            self._project.uuid, self._label._label_or_metadata, self._label.uuid
        )

    def _new_obj_from_dict(self, data):
        """Creates a new label from data in the dictionary.

        Args:
            data (dict): contains label_value properties value, uuid

        Returns:
            label object

        """

        obj = self._objclass(self._connection, self._project, self._label)
        obj.initialize_from_dict(data)
        return obj

    def __str__(self):
        s = ""
        if self._set:
            label = self._set[0]._label
            s = "LABEL\n"
            s += "\tname: " + str(label.name) + " uuid: " + str(label.uuid) + "\n"
            s += "LABEL VALUES\n"
            for lbv in self._set:
                s += "\tvalue: " + str(lbv.value) + " uuid:" + str(lbv.uuid) + "\n"

        return s



import datetime
import json
import os

import pandas as pd
from pandas import DataFrame

import sensiml.base.utility as utility
from sensiml.datamanager.capture_configurations import CaptureConfigurations
from sensiml.datamanager.captures import Captures, BackgroundCaptures
from sensiml.datamanager.capture import BackgroundCapture
from sensiml.datamanager.featurefiles import FeatureFiles
from sensiml.datamanager.knowledgepack import KnowledgePack
from sensiml.datamanager.queries import Queries
from sensiml.datamanager.sandboxes import Sandboxes
from sensiml.method_calls.functioncall import FunctionCall


class Project(object):
    """Base class for a project."""

    _uuid = ""
    _name = ""
    _created_at = None
    _schema = {}
    _settings = {}
    _query_optimized = True

    def __init__(self, connection):
        """Initialize a project instance.

        Args:
            connection (connection object)
        """
        self._connection = connection
        self._feature_files = FeatureFiles(self._connection, self)
        self._captures = Captures(self._connection, self)
        self._background_captures = BackgroundCaptures(self._connection)
        self._background_capture = BackgroundCapture(self._connection)
        self._sandboxes = Sandboxes(self._connection, self)
        self._queries = Queries(self._connection, self)
        self._plugin_config = None
        self._capture_configurations = CaptureConfigurations(self._connection, self)

    @property
    def uuid(self):
        """Auto generated unique identifier for the project object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """Name of the project object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def created_at(self):
        """Date of the project creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        if value:
            self._created_at = datetime.datetime.strptime(
                value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
            )

    @property
    def schema(self):
        """Schema of the project object"""
        return self._schema

    @schema.setter
    def schema(self, value):
        self._schema = value

    @property
    def plugin_config(self):
        """Plugin Config of the project object"""
        return self._plugin_config

    @plugin_config.setter
    def plugin_config(self, value):
        self._plugin_config = value

    @property
    def settings(self):
        """Global settings of the project object"""
        return self._settings

    @property
    def query_optimized(self):
        return self._query_optimized

    def add_segmenter(self, name, segmenter, preprocess=None, custom=False):
        """Saves a segmentation algorithm as the project's global segmentation setting.

        Args:
            name(str): Name to call the segmenter
            segmenter(FunctionCall): segmentation call object that the project will use by default
            preprocess(dict): Segment transforms to perform before segmenter
            custom(bool): a custom segmenter, or one of our server side segmenters


        """
        url = "project/{0}/segmenter/".format(self.uuid)
        if segmenter is not None:
            if not isinstance(segmenter, FunctionCall):
                print("segmenter is not a function call.")
                return
            segmenter_dict = segmenter._to_dict()
            if not segmenter_dict["type"] == "segmenter":
                print("segmenter is not a function call for a segmenter")
                return
            parameters = json.dumps(segmenter_dict)
        else:
            parameters = None

        if not isinstance(custom, bool):
            print("Custom must either be true or false.")
            return

        if preprocess:
            if isinstance(preprocess, dict):
                preprocess = json.dumps(preprocess)

        segmenter_info = {
            "custom": custom,
            "name": str(name),
            "parameters": parameters,
            "preprocess": preprocess,
        }

        request = self._connection.request("post", url, segmenter_info)
        response, err = utility.check_server_response(request)
        if err is False:
            print("Segmenter Uploaded.")

        return response

    def insert(self):
        """Calls the REST API to insert a new object, uses only the name and schema."""
        url = "project/"
        project_info = {
            "name": self.name,
            "capture_sample_schema": self.schema,
            "settings": self.settings,
            "plugin_config": self.plugin_config,
        }
        request = self._connection.request("post", url, project_info)
        response, err = utility.check_server_response(request)
        if err is False:
            self.uuid = response["uuid"]
            self._settings = response["settings"]
            self._query_optimized = response.get("optimized", True)
            self.plugin_config = response["plugin_config"]

        return response

    def update(self):
        """Calls the REST API to update the object."""
        url = "project/{0}/".format(self.uuid)
        project_info = {
            "name": self.name,
            "capture_sample_schema": self.schema,
            "settings": self.settings,
            "plugin_config": self.plugin_config,
        }
        request = self._connection.request("patch", url, project_info)
        response, err = utility.check_server_response(request)
        if err is False:
            self.plugin_config = response.get("plugin_config", None)

        return response

    def delete(self):
        """Calls the REST API to delete the object."""
        url = "project/{0}/".format(self.uuid)
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(request)

        return response

    def delete_async(self):
        """Calls the Async REST API to delete the project."""
        url = "project-delete/{0}/".format(self.uuid)
        request = self._connection.request("post", url)
        response, err = utility.check_server_response(request)

        return response

    def delete_async_status(self):
        """Calls the Async REST API to get the status of deleting the project."""
        url = "project-delete/{0}/".format(self.uuid)
        request = self._connection.request("get", url)
        response, err = utility.check_server_response(request)

        return response

    def delete_async_stop(self):
        """Calls the Async REST API to stop deleting the project."""
        url = "project-delete/{0}/".format(self.uuid)
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(request)

        return response

    def refresh(self):
        """Calls the REST API and self populates from the server."""
        url = "project/{0}/".format(self.uuid)
        request = self._connection.request("get", url)
        response, err = utility.check_server_response(request)
        if err is False:
            self.name = response["name"]
            self.schema = response["capture_sample_schema"]
            self._query_optimized = response.get("optimized", True)
            self.plugin_config = response.get("plugin_config", None)
            self.created_at = response.get("created_at", None)

        return response

    def query_optimize(self, force=False):
        self.refresh()
        """Calls the REST API and optimizes or re-optimizes the project for querying."""
        if not self.schema or not len(self._captures.get_captures()):
            print(
                "Cannot query optimize {} until there are uploaded captures.".format(
                    self._name
                )
                + "If data was uploaded, try client.project.refresh() followed by client.project.query_optimize()."
            )
        elif not self._query_optimized:
            print(
                "{} is not optimized for querying. Optimizing now...".format(self._name)
            )
            self._create_profile()
        elif force:
            print("Re-optimizing {} for querying now...".format(self._name))
            self._delete_profile()
            self._create_profile()

    def _create_profile(self):
        """Calls the REST API and creates a profile for optimized query times."""
        url = "project/{0}/profile/".format(self.uuid)
        request = self._connection.request("post", url)
        response, err = utility.check_server_response(request, is_octet=True)
        self.refresh()

        return response

    def _delete_profile(self):
        """Calls the REST API and drops the project profile."""
        url = "project/{0}/profile/".format(self.uuid)
        # Make a call to delete the profile
        request = self._connection.request("delete", url)
        response, err = utility.check_server_response(request, is_octet=True)
        self.refresh()

        return response

    def get_knowledgepack(self, kp_uuid):
        """Gets the KnowledgePack(s) created by the sandbox.

        Returns:
            a KnowledgePack instance, list of instances, or None
        """
        url = "knowledgepack/{0}/".format(kp_uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            kp = KnowledgePack(
                self._connection, self.uuid, response_data.get("sandbox_uuid")
            )
            kp.initialize_from_dict(response_data)
            return kp

    def _get_knowledgepacks(self):
        """Gets the KnowledgePack(s) created by the sandbox.

        Returns:
            a KnowledgePack instance, list of instances, or None
        """
        url = "project/{0}/knowledgepack/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return DataFrame(response_data)

    def list_knowledgepacks(self):
        """Lists all of the knowledgepacks associated with this project

        Returns:
            DataFrame: knowledgepacks on kb cloud
        """

        knowledgepacks = self._get_knowledgepacks().rename(
            columns={
                "name": "Name",
                "project_name": "Project",
                "sandbox_name": "Pipeline",
                "uuid": "kp_uuid",
                "created_at": "Created",
                "knowledgepack_description": "kp_description",
            }
        )

        if len(knowledgepacks) < 1:
            print("No Knowledgepacks stored for this project on the cloud.")
            return None
        return knowledgepacks[knowledgepacks["Name"] != ""][
            [
                "Name",
                "accuracy",
                "features_count",
                "Created",
                "Project",
                "Pipeline",
                "kp_uuid",
                "kp_description",
            ]
        ]

    def initialize_from_dict(self, dict):
        """Reads a json dict and populates a single project.

        Args:
            dict (dict): contains the project's 'name', 'uuid', 'schema', and 'settings' properties
        """
        self.uuid = dict["uuid"]
        self.name = dict["name"]
        self.schema = dict["capture_sample_schema"]
        self._settings = dict.get("settings", [])
        self._query_optimized = dict.get("optimized", True)
        self.plugin_config = dict.get("plugin_config", None)
        self.created_at = dict.get("created_at", None)

    def __getitem__(self, key):
        if type(key) is str:
            return self.captures.get_capture_by_filename(key)
        else:
            return self.captures.get_captures()[key]

    @property
    def featurefiles(self):
        return self._feature_files

    @property
    def captures(self):
        return self._captures

    @property
    def sandboxes(self):
        return self._sandboxes

    @property
    def queries(self):
        return self._queries

    @property
    def capture_configurations(self):
        return self._capture_configurations

    def columns(self):
        """Returns the sensor columns available in the project.

        Returns:
            columns (list[str]): a list of string names of the project's sensor columns
        """
        try:
            columnnames = self.schema.keys()
            return columnnames
        except:
            return None

    def metadata_columns(self):
        """Returns the metadata columns available in the project.

        Returns:
            columns (list[str]): a list of string names of the project's metadata columns
        """
        return self.captures.get_metadata_names()

    def metadata_values(self):
        return self.captures.get_metadata_names_and_values()

    def label_values(self):
        return self.captures.get_label_names_and_values()

    def label_columns(self):
        """Returns the label columns available in the project.

        Returns:
            columns (list[str]): a list of string names of the project's metadata columns
        """
        return self.captures.get_label_names()

    def statistics(self):
        """Gets all capture statistics for the project.

        Returns:
            DataFrame of capture statistics
        """
        url = "project/{0}/statistics/".format(self.uuid)
        data = {"events": False}
        response = self._connection.request("get", url, json=data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            data = utility.make_statistics_table(response_data)
            return data
        else:
            return response

    def get_active_pipelines(self):
        url = "project/{0}/active/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            if response_data:
                return response_data

        return response

    def get_segmenters(self):
        url = "project/{0}/segmenter/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            if response_data:
                return DataFrame(response_data).set_index("id")

        return response

    def get_project_summary(self):
        url = f"project-summary/{self.uuid}/"
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        return response_data

    def get_capture_stats(self):
        url = "project/{0}/capture-stats/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            if response_data:
                return DataFrame(response_data)

        return None

    def get_project_dcli(self):
        url = "project/{0}/dcli/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return response_data

        return response



import json
import os

from sensiml.datamanager.label import Label
from sensiml.datamanager.labelvalue import LabelValue
from sensiml.datamanager.metadata import Metadata
from sensiml.datamanager.metadata_relationship import MetadataRelationship
from sensiml.datamanager.metadata_value import MetadataValue
from sensiml.datamanager.segment import Segment
from sensiml.datamanager.segmenter import SegmenterSet
from sensiml.dclproj import DCLProject
from sensiml.datamanager.metadata import MetadataSet
from sensiml.datamanager.metadata_value import MetadataValueSet
from sensiml.datamanager.label import LabelSet
from sensiml.datamanager.labelvalue import LabelValueSet

color_list = [
    "#FF1D7DC4",
    "#FF58D963",
    "#FFF49534",
    "#FF7754AB",
    "#FFD14D49",
    "#FFC8CF4E",
    "#FF70A0B5",
    "#FF13631A",
    "#FF319476",
    "#FFB570AC",
    "#FF707070",
    "#FF421609",
    "#FF132AD4",
    "#FFC7AD1C",
    "#FFD47400",
    "#FF330C4A",
    "#FF00A2FF",
]


def upload_project(client, name: str, dclproj_path: str):
    if name in client.list_projects()["Name"].values:
        print("Project with this name already exists.")
        return

    client.project = name
    basedir = os.path.dirname(dclproj_path)
    capture_dir = os.path.join(basedir, "data")
    dclproj = DCLProject(path=dclproj_path)

    # TODO: plugin config
    # plugin_config = dclproj.get("ProjectCapturePluginConfig", None)
    # if plugin_config is not None:
    #    client.project.plugin_config = plugin_config
    #    print("Capture Config Found, Updating...")
    #    client.project.update()

    session_set = {}
    for _, session in dclproj.list_sessions().iterrows():
        if session["parameters"]:
            params = json.loads(session["parameters"])
            call = client.functions.create_function_call(params["name"])
            for k, v in params["inputs"].items():
                setattr(call, k, v)
        else:
            call = None

        new_segmenter = client.project.add_segmenter(
            session["name"],
            call,
            preprocess=session["preprocess"],
            custom=True if session["parameters"] else False,
        )
        session_set[session["id"]] = new_segmenter

    label_set = {}

    for _, item in dclproj._list_table_raw("Label").iterrows():
        if item["metadata"] == 0:
            label = Label(client._connection, client.project)
            label.name = item["name"]
            label.is_dropdown = True if item["is_dropdown"] else False
            label.insert()
            label_set[item.id] = label
        else:
            metadata = Metadata(client._connection, client.project)
            metadata.name = item["name"]
            metadata.is_dropdown = True if item["is_dropdown"] else False
            metadata.insert()
            label_set[item.id] = metadata

    label_value_set = {}

    for _, item in dclproj._list_table_raw("LabelValue").iterrows():
        if label_set[item["label"]].metadata == False:
            label_value = LabelValue(
                client._connection, client.project, label_set[item["label"]]
            )
            label_value.color = item["color"]
            label_value.value = item["value"]
            label_value.insert()
            label_value_set[item.id] = label_value
        else:
            metadata_value = MetadataValue(
                client._connection, client.project, label_set[item["label"]]
            )
            metadata_value.color = None
            metadata_value.value = item["value"]
            metadata_value.insert()
            label_value_set[item.id] = metadata_value

    for _, capture in dclproj.list_captures().iterrows():
        print(
            "Uploading Capture data, metadata, and labels for {}".format(
                capture["name"]
            )
        )

        capture_obj = client.project.captures.create_capture(
            os.path.basename(capture["name"]),
            os.path.join(capture_dir, capture["name"]),
        )

        # TODO: Bulk metadata creation
        for _, metadata_relationship in dclproj.get_capture_metadata(
            capture["name"], include_ids=True
        ).iterrows():
            metadata_value = label_value_set[metadata_relationship["label_value_id"]]
            metadata = label_set[metadata_relationship["label_id"]]
            metadata_relationship_obj = MetadataRelationship(
                client._connection,
                client.project,
                capture_obj,
                metadata,
                metadata_value,
            )
            metadata_relationship_obj.insert()

        # TODO: Bulk label creation
        for _, label_relationship in dclproj.list_capture_segments(
            capture["name"], include_ids=True
        ).iterrows():
            label_value = label_value_set[label_relationship["label_value_id"]]
            label = label_set[label_relationship["label_id"]]

            if not session_set.get(label_relationship["segmenter_id"], None):
                continue

            segmenter = session_set[label_relationship["segmenter_id"]]["id"]
            label_relationship_obj = Segment(
                client._connection,
                client.project,
                capture_obj,
                segmenter,
                label,
                label_value,
            )

            if (
                label_relationship["capture_sample_sequence_start"] < 0
                or label_relationship["capture_sample_sequence_end"]
                <= label_relationship["capture_sample_sequence_start"]
            ):
                print("Invalid Segment: Skipping")
                print(label_relationship)
                continue

            label_relationship_obj.sample_start = label_relationship[
                "capture_sample_sequence_start"
            ]
            label_relationship_obj.sample_end = label_relationship[
                "capture_sample_sequence_end"
            ]
            label_relationship_obj.insert()


def upload_project_dcli(client, name: str, dcli_path: str):

    client.project = name
    basedir = os.path.dirname(dcli_path)
    capture_dir = os.path.join(basedir)

    dcli = json.load(open(dcli_path, "r"))

    session_set = SegmenterSet(client._connection, client.project).to_dict()
    label_set = {}
    label_value_set = {}
    metadata_set = MetadataSet(client._connection, client.project).to_dict()
    metadata_value_set = {}
    for key, metadata in metadata_set.items():
        metadata_value_set[key] = MetadataValueSet(
            client._connection, client.project, metadata
        ).to_dict()

    label_set = LabelSet(client._connection, client.project).to_dict()
    label_value_set = {}
    for key, label in label_set.items():
        label_value_set[key] = LabelValueSet(
            client._connection, client.project, label
        ).to_dict()

    color_index = {}
    for index, capture_info in enumerate(dcli):

        if not os.path.exists(os.path.join(capture_dir, capture_info["file_name"])):
            print(f"Skipping: Capture {capture_info['file_name']} not found.")
            continue

        print(f"Uploading Capture {index}/{len(dcli)}: {capture_info['file_name']}")

        try:
            capture_obj = client.project.captures.create_capture(
                os.path.basename(capture_info["file_name"]),
                os.path.join(capture_dir, capture_info["file_name"]),
            )
        except Exception as e:
            print(e)
            continue

        if capture_info.get("metadata"):
            for metadata in capture_info["metadata"]:

                if not metadata["value"]:
                    continue

                metadata_obj = metadata_set.get(metadata["name"])

                if metadata_obj is None:
                    metadata_obj = Metadata(client._connection, client.project)
                    metadata_obj.name = metadata["name"]
                    metadata_obj.is_dropdown = True  # TODO check dcli if metadata_obj["is_dropdown"] else False
                    metadata_obj.insert()
                    metadata_set[metadata_obj.name] = metadata_obj
                    metadata_set[metadata_obj.name] = {}
                    metadata_set = MetadataSet(
                        client._connection, client.project
                    ).to_dict()
                    metadata_value_set[metadata_obj.name] = {}

                metadata_value_obj = metadata_value_set[metadata_obj.name].get(
                    metadata["value"]
                )

                if metadata_value_obj is None:
                    metadata_value_obj = MetadataValue(
                        client._connection, client.project, metadata_obj
                    )
                    metadata_value_obj.color = None
                    metadata_value_obj.value = metadata["value"]
                    metadata_value_obj.insert()

                    metadata_value_set[metadata_obj.name][
                        metadata_value_obj.value
                    ] = metadata_value_obj

                metadata_relationship_obj = MetadataRelationship(
                    client._connection,
                    client.project,
                    capture_obj,
                    metadata_obj,
                    metadata_value_obj,
                )
                metadata_relationship_obj.insert()

        if capture_info.get("sessions"):
            for session in capture_info["sessions"]:
                # TODO: Params
                # params = json.loads(session["parameters"])
                # call = client.functions.create_function_call(params["name"])
                # for k, v in params["inputs"].items():
                #    setattr(call, k, v)
                # else:
                # call = None

                if session["session_name"] not in session_set:
                    new_segmenter = client.project.add_segmenter(
                        session["session_name"],
                        segmenter=None,
                        preprocess=None,
                        custom=True,  # True if session["parameters"] else False,
                    )
                    session_set = SegmenterSet(
                        client._connection, client.project
                    ).to_dict()

                session_obj = session_set[session["session_name"]]

                for segment in session["segments"]:

                    label_obj = label_set.get(segment["name"])

                    if label_obj is None:
                        label_obj = Label(client._connection, client.project)
                        label_obj.name = segment["name"]
                        label_obj.is_dropdown = True
                        label_obj.insert()
                        label_set = LabelSet(
                            client._connection, client.project
                        ).to_dict()
                        label_value_set[label_obj.name] = {}
                        color_index[label_obj.name] = 0

                    if color_index.get(label_obj.name) is None:
                        color_index[label_obj.name] = 0

                    label_value_obj = label_value_set[label_obj.name].get(
                        segment["value"]
                    )
                    if label_value_obj is None:
                        label_value_obj = LabelValue(
                            client._connection, client.project, label_obj
                        )
                        # label_value_obj.color = segment["color"]
                        label_value_obj.value = segment["value"]

                        if color_index[label_obj.name] >= len(color_list):
                            color_index[label_obj.name] = 0

                        label_value_obj.color = color_list[color_index[label_obj.name]]
                        color_index[label_obj.name] += 1

                        label_value_obj.insert()

                        label_value_set[label_obj.name][
                            label_value_obj.value
                        ] = label_value_obj

                    label_relationship_obj = Segment(
                        client._connection,
                        client.project,
                        capture_obj,
                        session_obj.uuid,
                        label_obj,
                        label_value_obj,
                    )

                    if segment["end"] > capture_obj.capture_info["max_sequence"]:
                        segment["end"] = capture_obj.capture_info["max_sequence"]

                    if segment["start"] < 0:
                        segment["start"] = 0

                    label_relationship_obj.sample_start = segment["start"]
                    label_relationship_obj.sample_end = segment["end"]
                    try:
                        label_relationship_obj.insert()
                    except Exception as e:
                        print(e)



from warnings import warn


class PipelineWarning(Warning):
    """Base class for a pipeline warning"""

    pass


class PipelineError(Exception):
    """Base class for a pipeline error"""

    pass


class PipelineStep(object):
    """Base class for a pipeline step"""

    def __init__(self, name, step_type="default"):
        self._inputs = []
        self._outputs = []
        self._name = name
        self._step_type = step_type
        self._feature_table = None

    @property
    def feature_table(self):
        return self._feature_table

    @feature_table.setter
    def feature_table(self, value):
        self._feature_table = value

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        self._outputs = outputs

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def step_type(self):
        return self._step_type

    def _check_inputs(self, output_vars=()):
        """Compares the list of inputs with the given output_vars list.

        Args:
            output_vars (list[str]): strings representing the current set of output variables

        Returns:
            (list): unmatched inputs, or an empty list if all inputs are matched
        """
        unmatched_inputs = []
        for input in self.inputs:
            if not input in output_vars:
                unmatched_inputs.append(input)
        return unmatched_inputs

    def _check_outputs(self, output_vars=(), is_last=False):
        """Compares the outputs from the item with the other output_vars from the pipeline.

        Args:
            output_vars (list[str]): strings representing the current set of output variables
            is_last (bool): True if item is the last step, otherwise False (default)

        Returns:
            (list): duplicate outputs, or an empty list if all outputs are unique

        Raises:
            PipelineError if step is not the last one and does not have any outputs
            PipelineWarning if step is the last one and does not have any outputs
        """
        duplicate_outputs = []
        if self.outputs:
            for output in self.outputs:
                if output in output_vars:
                    duplicate_outputs.append(output)
        elif is_last:
            warn(
                "{0} {1} does not have any outputs.".format(
                    self._step_type, self._name
                ),
                PipelineWarning,
            )
        else:  # is_last = False
            raise PipelineError(
                "{0} {1} does not have any outputs.".format(self._step_type, self._name)
            )
        return duplicate_outputs



from uuid import UUID
import sensiml.base.utility as utility
from sensiml.datamanager.base import Base, BaseSet


class Segmenter(Base):
    """Base class for a segmenter object."""

    _data = None
    _fields = [
        "uuid",
        "name",
        "parameters",
        "preprocess",
        "custom",
        "parent",
        "last_modified",
        "created_at",
    ]
    _field_map = {
        "uuid": "id",
    }

    def __init__(self, connection, project):
        """Initialize a metadata object.

        Args:
            connection
            project
        """
        self._uuid = ""
        self._name = ""
        self._parameters = None
        self._preprocess = None
        self._custom = True
        self._parent = None
        self._last_modified = ""
        self._created_at = ""
        self._connection = connection
        self._project = project

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        """The name property of the metadata object"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    @property
    def preprocess(self):
        return self._preprocess

    @preprocess.setter
    def preprocess(self, value):
        self._preprocess = value

    @property
    def custom(self):
        return self._custom

    @custom.setter
    def custom(self, value):
        if not isinstance(value, bool):
            raise ValueError("custom must be a bool.")

        self._custom = value

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        if value is None:
            self._parent = None
        else:
            self._parent = UUID(value)

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/segmenter/".format(self._project.uuid)

        data = self._to_representation()

        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)

        if err is False:
            self.uuid = response_data["id"]

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""
        url = "project/{0}/segmenter/{1}/".format(self._project.uuid, self.uuid)

        data = self._to_representation()

        response = self._connection.request("put", url, data)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/segmenter/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/segmenter/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response


class SegmenterSet(BaseSet):
    """Base class for a segmenter object."""

    def __init__(self, connection, project, initialize_set=True):
        self._connection = connection
        self._project = project
        self._set = None
        self._objclass = Segmenter
        self._attr_key = "name"

        if initialize_set:
            self.refresh()

    @property
    def get_set_url(self):
        return "project/{0}/segmenter/".format(self._project.uuid)



import warnings


class DeviceConfig(dict):
    """
    Defines dictionary class for device configuration to be sent from sensiml to the
    Server for KnowledgePack code/binary generation.
    """

    # List of fields and their defaults
    fields = {
        "target_platform": 0,
        "build_flags": "",
        "budget": {},
        "debug": False,
        "sram_size": None,
        "test_data": "",
        "application": "",
        "sample_rate": 100,
        "kb_description": None,
    }

    def __init__(self, iterable=None, strict=True, **kwargs):
        super(DeviceConfig, self).__init__()
        object.__setattr__(self, "_strict", strict)
        for k, v in self.fields.items():
            self[k] = v
        if iterable:
            for k, v in dict(iterable).items():
                setattr(self, k, v)
        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if self._strict and name not in self.fields.keys():
            out_str = "{} not in device configuration.".format(name)
            out_str += "\nYou may need to update the device configuration on KB Cloud."
            warnings.warn(out_str, DeprecationWarning)

            return
        if name == "budget":
            assert isinstance(value, dict), "Budget is not a dictionary."
        elif name == "debug":
            value = bool(value)

        self[name] = value

    def initialize_from_dict(self, init_dict):
        """DEPRECATED: Populates a single device config from a dictionary of properties.

        Args:
            (dict): contains target_platform, platform_version, build_flags, budget, debug and sram_size
        """
        warnings.warn(
            "initialize_from_dict() is deprecated. Please pass arguments to constructor",
            DeprecationWarning,
        )
        for k, v in init_dict.items():
            setattr(self, k, v)



from pandas import DataFrame
import numpy as np
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import scipy
from scipy import fftpack
from scipy import signal

from scipy.sparse import coo_matrix

from sensiml.datamanager.label import LabelSet
from sensiml.datamanager.labelvalue import LabelValueSet
from sensiml.datamanager.segment import Segment
from sensiml.datamanager.segmenter import SegmenterSet


def upload_datasegments(client, datasegments, default_label="Unknown", verbose=True):
    """Upload a segments form DataSegments object

    Args:
        client : logged in Client set to a particular project
        datasegments (DataSegments): DataSegments object containing segments to upload
        default_label (str, optional): Label to use as unknown if DataSegments label is not part of the cloud project. Defaults to "Unknown".
    """

    ls = LabelSet(client._connection, client.project)
    # TODO: Support multiple label groups
    # lvs = {label.name:LabelValueSet(client._connection, client.project, label) for label is ls.labels}
    label = ls.labels[0]
    lvs = LabelValueSet(client._connection, client.project, label).to_dict()
    segmenter_set = SegmenterSet(client._connection, client.project).to_dict()

    if lvs.get(default_label) is None:
        print("Default label is not part of label set")
        return

    default_label_value = lvs[default_label]

    for index, datasegment in enumerate(datasegments):
        capture = client.get_captures()[datasegment.capture]
        label_value = lvs.get(datasegment.label_value, default_label_value)
        segmenter = segmenter_set[datasegment.session]
        tmp_segment = Segment(
            client._connection,
            client.project,
            capture,
            segmenter=segmenter,
            label=label,
            label_value=label_value,
        )
        tmp_segment.uuid = datasegment.uuid
        tmp_segment.sample_start = datasegment.start
        tmp_segment.sample_end = datasegment.end
        if datasegment.uuid:
            tmp_segment.update()
        else:
            tmp_segment.insert()
        if verbose:
            print(capture.filename, index, tmp_segment.uuid)


def confusion_matrix(y_true, y_pred, labels, normalize=None):

    labels = np.asarray(labels)
    n_labels = labels.size
    if n_labels == 0:
        raise ValueError("'labels' should contains at least one label.")
    elif y_true.size == 0:
        return np.zeros((n_labels, n_labels), dtype=int)
    elif len(np.intersect1d(y_true, labels)) == 0:
        raise ValueError("At least one label specified must be in y_true")

    sample_weight = np.ones(y_true.shape[0], dtype=np.int64)

    # check_consistent_length(y_true, y_pred, sample_weight)

    if normalize not in ["true", "pred", "all", None]:
        raise ValueError("normalize must be one of {'true', 'pred', 'all', None}")

    n_labels = labels.size
    # If labels are not consecutive integers starting from zero, then
    # y_true and y_pred must be converted into index form
    need_index_conversion = not (
        labels.dtype.kind in {"i", "u", "b"}
        and np.all(labels == np.arange(n_labels))
        and y_true.min() >= 0
        and y_pred.min() >= 0
    )
    if need_index_conversion:
        label_to_ind = {y: x for x, y in enumerate(labels)}
        y_pred = np.array([label_to_ind.get(x, n_labels + 1) for x in y_pred])
        y_true = np.array([label_to_ind.get(x, n_labels + 1) for x in y_true])

    # intersect y_pred, y_true with labels, eliminate items not in labels
    ind = np.logical_and(y_pred < n_labels, y_true < n_labels)
    if not np.all(ind):
        y_pred = y_pred[ind]
        y_true = y_true[ind]
        # also eliminate weights of eliminated items
        sample_weight = sample_weight[ind]

    # Choose the accumulator dtype to always have high precision
    dtype = np.int64

    cm = coo_matrix(
        (sample_weight, (y_true, y_pred)),
        shape=(n_labels, n_labels),
        dtype=dtype,
    ).toarray()

    with np.errstate(all="ignore"):
        if normalize == "true":
            cm = cm / cm.sum(axis=1, keepdims=True)
        elif normalize == "pred":
            cm = cm / cm.sum(axis=0, keepdims=True)
        elif normalize == "all":
            cm = cm / cm.sum()
        cm = np.nan_to_num(cm)

    return cm


def compute_confusion_matrix(
    predicted, ground_truth, ytick_labels=None, ground_truth_unlabled_is_unknown=False
) -> dict:

    if ytick_labels is None:
        ytick_labels = []

    df_gt = DataFrame(
        [
            "Unlabeled"
            for _ in range(
                max(
                    ground_truth[-1]["capture_sample_sequence_end"],
                    predicted[-1]["capture_sample_sequence_end"],
                )
                + 1
            )
        ],
        columns=["label_value"],
    )

    for segment in ground_truth:

        df_gt.iloc[
            segment["capture_sample_sequence_start"] : segment[
                "capture_sample_sequence_end"
            ]
        ] = segment["label_value"]

    for segment in predicted:
        # print(segment["capture_sample_sequence_start"])
        # print(segment["capture_sample_sequence_end"])
        segment["y_predicted"] = df_gt.iloc[
            segment["capture_sample_sequence_start"] : segment[
                "capture_sample_sequence_end"
            ]
        ]["label_value"].mode()[0]

    combined = DataFrame(predicted)
    if ground_truth_unlabled_is_unknown == False:
        combined = combined[combined["y_predicted"] != "Unlabeled"]

    # TODO: Discuss a better way to handle this situation,  for now I think the best method is to return confusion matrix
    # if not set(combined["Label_Value"].values) <= set(ytick_labels):
    #    return get_empty_confusion_matrix(ytick_labels)

    ytick_labels = list(
        set(combined["label_value"].values.astype(str)).union(ytick_labels)
    )

    ytick_labels = sorted(
        list(set(ytick_labels).union(set(combined["y_predicted"].values.astype(str))))
    )

    cm_array = confusion_matrix(
        y_true=combined["label_value"].astype(str),
        y_pred=combined["y_predicted"].astype(str),
        labels=ytick_labels,
    )

    df_cm = DataFrame(cm_array, index=ytick_labels, columns=ytick_labels).T

    df_cm["Total"] = df_cm.T.sum(axis=0)

    return df_cm


def plot_segments_labels(
    segments,
    data=None,
    labels=None,
    figsize=(30, 4),
    title=None,
    axes=None,
    y_label=(0, 1),
):

    if hasattr(segments, "to_dataframe"):
        segments = segments.to_dataframe()

    if axes is None:
        plt.figure(figsize=figsize)
        currentAxis = plt.gca()
    else:
        currentAxis = axes

    cmap = plt.cm.rainbow

    if labels is None:
        labels = sorted(segments.label_value.unique().tolist())

    if not labels:
        return

    if data is not None:
        y_label = [data.min(), data.max() - data.min()]

    delta = 1 / len(labels)
    label_float = np.arange(0, 1, delta)
    label_float[-1] = 1.0

    label_colors = {labels[index]: cmap(x) for index, x in enumerate(label_float)}
    label_legend = [Line2D([0], [0], color=cmap(x), lw=4) for x in label_float] + [
        Line2D([0], [0], color="white", lw=4)
    ]

    x_lim_end = 0
    for _, seg in segments.iterrows():
        y_origin = y_label[0]
        x_origin = seg["capture_sample_sequence_start"]
        x_final = (
            seg["capture_sample_sequence_end"] - seg["capture_sample_sequence_start"]
        )
        y_final = y_label[1]

        currentAxis.add_artist(
            plt.Rectangle(
                (x_origin, y_origin),
                x_final,
                y_final,
                alpha=0.7,
                color=label_colors[seg["label_value"]],
            )
        )
        x_lim_end = seg["capture_sample_sequence_end"]

    currentAxis.legend(label_legend, labels + [""], loc=1)
    currentAxis.set_xlim(0, x_lim_end)
    currentAxis.set_title(title)

    if data is not None:
        currentAxis.plot(data)


def add_eps(x):
    x[np.where(x == 0)] = scipy.finfo(dtype=x.dtype).eps
    return x


def preemphasis(seq, coeff):
    return np.append(seq[0], seq[1:] - coeff * seq[:-1])


# http://www.practicalcryptography.com/miscellaneous/machine-learning/guide-mel-frequency-cepstral-coefficients-mfccs/
def freq_to_mel(freq):
    return 1125.0 * np.lib.scimath.log(1.0 + freq / 700.0)


def mel_to_freq(mel):
    return 700.0 * (np.exp(mel / 1125.0) - 1.0)


def iter_bin(out, curr_bin, next_bins, backward=False):
    next_bin = next_bins[np.where(next_bins > curr_bin)][0]
    if backward:
        sign = -1
        bias = next_bin
    else:
        sign = 1
        bias = curr_bin
    for f in range(int(curr_bin), int(next_bin)):
        out[f] = sign * (f - bias) / (next_bin - curr_bin)


def mel_filterbank(num_bank, num_freq, sample_freq, low_freq, high_freq):
    num_fft = (num_freq - 1) * 2
    low_mel = freq_to_mel(low_freq)
    high_mel = freq_to_mel(high_freq)
    banks = np.linspace(low_mel, high_mel, num_bank + 2)
    bins = np.floor((num_fft + 1) * mel_to_freq(banks) / sample_freq)
    out = np.zeros((num_bank, num_fft // 2 + 1))
    for b in range(num_bank):
        iter_bin(out[b], bins[b], bins[b + 1 :])
        iter_bin(out[b], bins[b + 1], bins[b + 2 :], backward=True)
    return out


def mfcc(
    data,
    channel,
    sample_freq,
    frame_length=1024,
    frame_shift=0.5,
    remove_dc_offset=True,
    num_ceps=23,
    num_mel_bins=80,
    use_power=True,
    figsize=(30, 4),
    plot_all=True,
):
    """
    data : DataFrame data to use
    channel: channel (column) to use from dataframe
    sampel_rate: the sampel rate of the captured data. The sample rate defines the MFCC
    """

    def add_eps(x):
        x[np.where(x == 0)] = scipy.finfo(dtype=x.dtype).eps
        return x

    def preemphasis(seq, coeff):
        return np.append(seq[0], seq[1:] - coeff * seq[:-1])

    # http://www.practicalcryptography.com/miscellaneous/machine-learning/guide-mel-frequency-cepstral-coefficients-mfccs/
    def freq_to_mel(freq):
        return 1125.0 * np.lib.scimath.log(1.0 + freq / 700.0)

    def mel_to_freq(mel):
        return 700.0 * (np.exp(mel / 1125.0) - 1.0)

    def iter_bin(out, curr_bin, next_bins, backward=False):
        next_bin = next_bins[np.where(next_bins > curr_bin)][0]
        if backward:
            sign = -1
            bias = next_bin
        else:
            sign = 1
            bias = curr_bin
        for f in range(int(curr_bin), int(next_bin)):
            out[f] = sign * (f - bias) / (next_bin - curr_bin)

    def mel_filterbank(num_bank, num_freq, sample_freq, low_freq, high_freq):
        num_fft = (num_freq - 1) * 2
        low_mel = freq_to_mel(low_freq)
        high_mel = freq_to_mel(high_freq)
        banks = np.linspace(low_mel, high_mel, num_bank + 2)
        bins = np.floor((num_fft + 1) * mel_to_freq(banks) / sample_freq)
        out = np.zeros((num_bank, num_fft // 2 + 1))
        for b in range(num_bank):
            iter_bin(out[b], bins[b], bins[b + 1 :])
            iter_bin(out[b], bins[b + 1], bins[b + 2 :], backward=True)
        return out

    window_type = "hamming"

    # Fbank conf
    preemphasis_coeff = 0.97
    high_freq = 0.0  # offset from Nyquist freq [Hz]
    low_freq = 20.0  # offset from 0 [Hz]
    num_lifter = 23

    raw_seq = data[channel].values
    assert raw_seq.ndim == 1  # assume mono
    seq = raw_seq.astype(scipy.float64)
    if remove_dc_offset:
        seq -= np.mean(seq)

    # STFT feat
    seq = preemphasis(seq, preemphasis_coeff)
    window = signal.get_window(window_type, frame_length)
    mode = "psd" if use_power else "magnitude"
    f, t, spectrogram = signal.spectrogram(
        seq,
        sample_freq,
        window=window,
        noverlap=int(frame_length * frame_shift),
        mode=mode,
    )

    # log-fbank feat
    banks = mel_filterbank(
        num_mel_bins,
        spectrogram.shape[0],
        sample_freq,
        low_freq,
        sample_freq // 2 - high_freq,
    )

    fbank_spect = np.dot(banks, spectrogram)
    logfbank_spect = np.log(add_eps(fbank_spect))

    # mfcc feat
    dct_feat = fftpack.dct(logfbank_spect, type=2, axis=0, norm="ortho")[:num_ceps]
    lifter = 1 + num_lifter / 2.0 * np.sin(scipy.pi * np.arange(num_ceps) / num_lifter)

    mfcc_feat = lifter[:, scipy.newaxis] * dct_feat

    if plot_all:
        fig, axes = plt.subplots(nrows=4, ncols=1, figsize=(figsize[0], figsize[1] * 4))
        fig.tight_layout()
        axes[0].plot(seq)
        axes[0].set_title("signal")
        axes[1].imshow(np.log(spectrogram), aspect="auto")
        axes[1].set_title("log(spectrogram)")
        axes[2].matshow(logfbank_spect, aspect="auto")
        axes[2].set_title("logfbank")
        axes[3].matshow(mfcc_feat, aspect="auto")
        axes[3].set_title("mfcc")

        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(figsize[0], figsize[1]))

        for bank in banks:
            axes[0].plot(bank)
            axes[0].set_title("Filter banks")

        axes[1].plot(lifter)
        axes[1].set_title("Litering Coefficients")

        plt.subplots_adjust(hspace=0.2)

    return mfcc_feat



from sensiml.client import Client
from sensiml.client import Client as SensiML
from sensiml.client import __version__

__all__ = ["SensiML", "Client"]


try:
    from IPython.core.display import HTML

    display(HTML("<style>.container { width:90% !important; }</style>"))
except:
    pass

name = "sensiml"



class DataManagerError(Exception):
    pass


class AsyncObjectError(DataManagerError):
    pass


class AsyncObjectNotReadyError(AsyncObjectError):
    pass


class CaptureNotReadyError(AsyncObjectNotReadyError):
    pass


class AsyncObjectFailureError(AsyncObjectError):
    """When object processing/upload fails"""

    pass


class CaptureUploadFailureError(AsyncObjectFailureError):
    pass


# Renamed to event


class EventNotReadyError(CaptureNotReadyError):
    """Alias for CaptureNotReadyError since Captures
    are called Events in some locations."""

    pass


class EventUploadFailureError(CaptureUploadFailureError):
    pass
